from kooplearn.datasets.deterministic import DuffingOscillator, Lorenz63
from kooplearn.datasets.stochastic import (
    LangevinTripleWell1D,
    LinearModel,
    LogisticMap,
    Mock,
    MullerBrownPotential,
)



import logging
from typing import Optional

import numpy as np
from scipy.linalg import cho_factor, cho_solve, eig, eigh, lstsq, qr
from scipy.sparse.linalg import eigs, eigsh
from sklearn.utils.extmath import randomized_svd

from kooplearn._src.linalg import eigh_rank_reveal, weighted_norm
from kooplearn._src.utils import fuzzy_parse_complex, topk

logger = logging.getLogger("kooplearn")


def postprocess_UV(U, V, rank):
    assert U.shape == V.shape
    if U.shape[1] < rank:
        logger.warning(
            f"Warning: The fitting algorithm discarded {rank - U.shape[1]} dimensions of the {rank} requested out of numerical instabilities.\nThe rank attribute has been updated to {U.shape[1]}.\nConsider decreasing the rank parameter."
        )
    else:
        # Assuming that everything is in decreasing order
        U = U[:, :rank]
        V = V[:, :rank]
    return U, V


def regularize(M: np.ndarray, reg: float):
    """Regularize a matrix by adding a multiple of the identity matrix to it.
    Args:
        M (np.ndarray): Matrix to regularize.
        reg (float): Regularization parameter.
    Returns:
        np.ndarray: Regularized matrix.
    """
    return M + (M.shape[0] * reg) * np.identity(M.shape[0], dtype=M.dtype)


def add_diagonal(M: np.ndarray, alpha: float):
    np.fill_diagonal(M, M.diagonal() + alpha)


# Reduced Rank Algorithms
def _filter_reduced_rank_svals(values, vectors, rank):
    eps = 2 * np.finfo(vectors.dtype).eps
    # Filtering procedure.
    # Create a mask which is True when the real part of the eigenvalue is negative or the imaginary part is nonzero
    is_invalid = np.logical_or(np.abs(np.real(values)) <= eps, np.imag(values) != 0)
    # Check if any is invalid take the first occurrence of a True value in the mask and filter everything after that
    if np.any(is_invalid):
        values = values[~is_invalid].real
        vectors = vectors[:, ~is_invalid]

    sort_perm = topk(values, len(values)).indices
    values = values[sort_perm]
    vectors = vectors[:, sort_perm]

    # Assert that the eigenvectors do not have any imaginary part
    assert np.all(
        np.imag(vectors) == 0
    ), "The eigenvectors should be real. Decrease the rank or increase the regularization strength."

    # Take the real part of the eigenvectors
    vectors = np.real(vectors)
    values = np.real(values)
    return values, vectors


def fit_reduced_rank_regression(
    kernel_X: np.ndarray,  # Kernel matrix of the input data
    kernel_Y: np.ndarray,  # Kernel matrix of the output data
    tikhonov_reg: float,  # Tikhonov regularization parameter, can be 0
    rank: int,  # Rank of the estimator
    svd_solver: str = "arnoldi",  # SVD solver to use. 'arnoldi' is faster but might be numerically unstable.
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    if tikhonov_reg == 0.0:
        return _fit_reduced_rank_regression_noreg(
            kernel_X, kernel_Y, rank, svd_solver=svd_solver
        )
    else:
        n_pts = kernel_X.shape[0]
        eps = 1000.0 * np.finfo(kernel_X.dtype).eps
        penalty = max(eps, tikhonov_reg) * n_pts
        A = np.multiply(kernel_Y, n_pts ** (-0.5)) @ np.multiply(
            kernel_X, n_pts ** (-0.5)
        )

        add_diagonal(kernel_X, penalty)
        # Find U via Generalized eigenvalue problem equivalent to the SVD. If K is ill-conditioned might be slow.
        # Prefer svd_solver == 'randomized' in such a case.
        if svd_solver == "arnoldi":
            # Adding a small buffer to the Arnoldi-computed eigenvalues.
            _num_arnoldi_eigs = min(rank + 5, A.shape[0])
            values, vectors = eigs(A, k=_num_arnoldi_eigs, M=kernel_X)
        else:  # 'full'
            values, vectors = eig(A, kernel_X, overwrite_a=True, overwrite_b=True)
        add_diagonal(kernel_X, -penalty)

        values, vectors = _filter_reduced_rank_svals(values, vectors, rank)
        # Compare the filtered eigenvalues with the regularization strength, and warn if there are any eigenvalues that are smaller than the regularization strength.
        if not np.all(np.abs(values) >= tikhonov_reg):
            logger.warning(
                f"Warning: {(np.abs(values) < tikhonov_reg).sum()} out of the {len(values)} squared singular values are smaller than the regularization strength {tikhonov_reg:.2e}. Consider redudcing the regularization strength to avoid overfitting."
            )
        # Eigenvector normalization
        kernel_X_vecs = np.dot(np.multiply(kernel_X, n_pts ** (-0.5)), vectors)
        vecs_norms = np.sum(
            kernel_X_vecs**2
            + tikhonov_reg * kernel_X_vecs * vectors * (n_pts**0.5),
            axis=0,
        ) ** (0.5)
        U = vectors / vecs_norms
        # Ordering the results
        sort_perm = topk(values, len(values)).indices
        U = U[:, sort_perm]
        V = kernel_X @ U
        svals = np.flip(np.sort(np.abs(values))) ** 0.5
        U, V = postprocess_UV(U, V, rank)
        return U, V, svals


def _fit_reduced_rank_regression_noreg(
    kernel_X: np.ndarray,  # Kernel matrix of the input data
    kernel_Y: np.ndarray,  # Kernel matrix of the output data
    rank: int,  # Rank of the estimator
    svd_solver: str = "arnoldi",  # Solver for the generalized eigenvalue problem. 'arnoldi' or 'full'
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    # Solve the Hermitian eigenvalue problem to find V
    logger.warning(
        "The least-squares solution (tikhonov_reg == 0) of the reduced rank problem in the kernel setting is computationally very inefficient. Consider adding a small regularization parameter."
    )

    values_X, U_X = eigh(kernel_X)
    U_X, _, _ = eigh_rank_reveal(values_X, U_X, kernel_X.shape[0], verbose=False)
    proj_X = U_X @ U_X.T
    L = proj_X @ kernel_Y
    if svd_solver != "full":
        values, vectors = eigs(L, rank + 3)
    else:
        values, vectors = eig(L, overwrite_a=True)
    values, V = _filter_reduced_rank_svals(values, vectors, rank)
    # Normalize V
    _V_norm = np.linalg.norm(V, ord=2, axis=0) / np.sqrt(V.shape[0])
    eps = 1000.0 * np.finfo(kernel_X.dtype).eps * V.shape[0]
    _inv_V_norm = np.where(_V_norm < eps, 0.0, _V_norm**-1)
    V = V / _inv_V_norm
    # Solve the least squares problem to determine U
    U = lstsq(kernel_X, V)[0]
    svals = np.flip(np.sort(np.abs(values))) ** 0.5
    U, V = postprocess_UV(U, V, rank)
    return U, V, svals


def fit_nystroem_reduced_rank_regression(
    kernel_X: np.ndarray,  # Kernel matrix of the input inducing points
    kernel_Y: np.ndarray,  # Kernel matrix of the output inducing points
    kernel_Xnys: np.ndarray,  # Kernel matrix between the input data and the input inducing points
    kernel_Ynys: np.ndarray,  # Kernel matrix between the output data and the output inducing points
    tikhonov_reg: float = 0.0,  # Tikhonov regularization parameter
    rank: Optional[int] = None,  # Rank of the estimator
    svd_solver: str = "arnoldi",  # Solver for the generalized eigenvalue problem. 'arnoldi' or 'full'
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    dim = kernel_Xnys.shape[0]
    num_centers = kernel_X.shape[0]

    eps = 1000 * np.finfo(kernel_X.dtype).eps * num_centers
    reg = max(eps, tikhonov_reg)

    # LHS of the generalized eigenvalue problem
    rsqrt_Mn = (num_centers * dim) ** -0.5
    kernel_YX_nys = (rsqrt_Mn * kernel_Ynys.T) @ (rsqrt_Mn * kernel_Xnys)

    _tmp_YX = lstsq(kernel_Y * (num_centers**-1), kernel_YX_nys)[0]
    kernel_XYX = kernel_YX_nys.T @ _tmp_YX

    # RHS of the generalized eigenvalue problem
    kernel_Xnys_sq = (rsqrt_Mn * kernel_Xnys.T) @ (
        kernel_Xnys * rsqrt_Mn
    ) + reg * kernel_X * (num_centers**-1)

    add_diagonal(kernel_Xnys_sq, eps)
    A = lstsq(kernel_Xnys_sq, kernel_XYX)[0]
    if svd_solver == "full":
        values, vectors = eigh(
            kernel_XYX, kernel_Xnys_sq
        )  # normalization leads to needing to invert evals
    elif svd_solver == "arnoldi":
        _oversampling = max(10, 4 * int(np.sqrt(rank)))
        _num_arnoldi_eigs = min(rank + _oversampling, kernel_X.shape[0])
        values, vectors = eigs(kernel_XYX, k=_num_arnoldi_eigs, M=kernel_Xnys_sq)
    else:
        raise ValueError(f"Unknown svd_solver {svd_solver}")
    add_diagonal(kernel_Xnys_sq, -eps)

    values, vectors = _filter_reduced_rank_svals(values, vectors, rank)
    # Compare the filtered eigenvalues with the regularization strength, and warn if there are any eigenvalues that are smaller than the regularization strength.
    if not np.all(np.abs(values) >= tikhonov_reg):
        logger.warning(
            f"Warning: {(np.abs(values) < tikhonov_reg).sum()} out of the {len(values)} squared singular values are smaller than the regularization strength {tikhonov_reg:.2e}. Consider redudcing the regularization strength to avoid overfitting."
        )
    # Eigenvector normalization
    normalization_csts = (
        np.abs(np.sum(vectors.conj() * (kernel_XYX @ vectors), axis=0))
    ) ** 0.5
    has_small_norm = normalization_csts < 1000.0 * np.finfo(vectors.dtype).eps
    vectors = vectors[:, ~has_small_norm] / normalization_csts[~has_small_norm]
    # Ordering the results
    sort_perm = topk(values[~has_small_norm], np.sum(~has_small_norm)).indices
    vectors = vectors[:, sort_perm]
    U = A @ vectors
    V = _tmp_YX @ vectors
    svals = np.flip(np.sort(np.abs(values))) ** 0.5
    U, V = postprocess_UV(U, V, rank)
    return U.real, V.real, svals


def fit_rand_reduced_rank_regression(
    kernel_X: np.ndarray,  # Kernel matrix of the input data
    kernel_Y: np.ndarray,  # Kernel matrix of the output data
    tikhonov_reg: float,  # Tikhonov regularization parameter
    rank: int,  # Rank of the estimator
    n_oversamples: int,  # Number of oversamples
    optimal_sketching: bool,  # Whether to use optimal sketching (slower but more accurate) or not.
    iterated_power: int,  # Number of iterations of the power method
    rng_seed: Optional[
        int
    ] = None,  # Seed for the random number generator (for reproducibility)
    precomputed_cholesky=None,  # Precomputed Cholesky decomposition. Should be the output of cho_factor evaluated on the regularized kernel matrix.
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    rng = np.random.default_rng(rng_seed)
    dim = kernel_X.shape[0]
    inv_dim = dim ** (-1.0)

    penalty = dim * tikhonov_reg
    add_diagonal(kernel_X, penalty)
    if precomputed_cholesky is None:
        cholesky_decomposition = cho_factor(kernel_X)
    else:
        cholesky_decomposition = precomputed_cholesky
    add_diagonal(kernel_X, -penalty)

    sketch_dimension = rank + n_oversamples

    if optimal_sketching:
        Cov = inv_dim * kernel_Y
        sketch = rng.multivariate_normal(
            np.zeros(dim, dtype=kernel_Y.dtype), Cov, size=sketch_dimension
        ).T
    else:
        sketch = rng.standard_normal(size=(dim, sketch_dimension))

    for _ in range(iterated_power):
        # Powered randomized rangefinder
        sketch = (inv_dim * kernel_Y) @ (
            sketch - penalty * cho_solve(cholesky_decomposition, sketch)
        )
        sketch, _ = qr(sketch, mode="economic")  # QR re-orthogonalization

    kernel_X_sketch = cho_solve(cholesky_decomposition, sketch)
    _M = sketch - penalty * kernel_X_sketch

    F_0 = sketch.T @ sketch - penalty * (sketch.T @ kernel_X_sketch)  # Symmetric
    F_0 = 0.5 * (F_0 + F_0.T)
    F_1 = _M.T @ (inv_dim * (kernel_Y @ _M))

    values, vectors = eig(lstsq(F_0, F_1)[0])

    values, vectors = _filter_reduced_rank_svals(values, vectors, rank)
    # Remove elements in the kernel of F_0
    _threshold = 1000.0 * np.finfo(vectors.dtype).eps * vectors.shape[0]
    relative_norm_sq = (
        np.sum(vectors.conj() * (F_0 @ vectors), axis=0)
        / np.linalg.norm(vectors, axis=0) ** 2
    )
    is_in_kernel = np.abs(relative_norm_sq) < _threshold
    num_in_kernel = is_in_kernel.sum()
    if num_in_kernel > 0:
        # Logging. Print the number of discarded eigenvalues.
        values = values[~is_in_kernel]
        vectors = vectors[:, ~is_in_kernel]

    normalization_csts = (np.sum(vectors.conj() * (F_0 @ vectors), axis=0).real) ** 0.5
    vectors = vectors / normalization_csts

    sort_perm = topk(values, len(values)).indices
    vectors = vectors[:, sort_perm]

    U = (dim**0.5) * kernel_X_sketch @ vectors
    V = (dim**0.5) * _M @ vectors

    svals = np.flip(np.sort(values)) ** 0.5
    U, V = postprocess_UV(U, V, rank)
    return U.real, V.real, svals


# Principal Component Algorithms
def fit_principal_component_regression(
    kernel_X: np.ndarray,  # Kernel matrix of the input data
    tikhonov_reg: float = 0.0,  # Tikhonov regularization parameter, can be zero
    rank: Optional[int] = None,  # Rank of the estimator
    svd_solver: str = "arnoldi",  # Solver for the generalized eigenvalue problem. 'arnoldi' or 'full'
) -> tuple[np.ndarray, np.ndarray]:
    dim = kernel_X.shape[0]
    add_diagonal(kernel_X, dim * tikhonov_reg)
    if svd_solver == "arnoldi":
        _num_arnoldi_eigs = min(rank + 5, kernel_X.shape[0])
        values, vectors = eigsh(kernel_X, k=_num_arnoldi_eigs)
    elif svd_solver == "full":
        values, vectors = eigh(kernel_X)
    else:
        raise ValueError(f"Unknown svd_solver {svd_solver}")
    add_diagonal(kernel_X, -dim * tikhonov_reg)

    vectors, values, rsqrt_values = eigh_rank_reveal(values, vectors, rank)
    Q = np.sqrt(dim) * vectors * (rsqrt_values)
    kernel_X_eigvalsh = (np.flip(np.sort(np.abs(values))) / dim) ** 0.5
    Q, Q = postprocess_UV(Q, Q, rank)
    return Q, Q, kernel_X_eigvalsh


def fit_nystroem_principal_component_regression(
    kernel_X: np.ndarray,  # Kernel matrix of the input inducing points
    kernel_Y: np.ndarray,  # Kernel matrix of the output inducing points
    kernel_Xnys: np.ndarray,  # Kernel matrix between the input data and the input inducing points
    kernel_Ynys: np.ndarray,  # Kernel matrix between the output data and the output inducing points
    tikhonov_reg: float = 0.0,  # Tikhonov regularization parameter (can be 0)
    rank: Optional[int] = None,  # Rank of the estimator
    svd_solver: str = "arnoldi",  # Solver for the generalized eigenvalue problem. 'arnoldi' or 'full'
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    dim = kernel_X.shape[0]
    eps = 1000 * np.finfo(kernel_X.dtype).eps
    reg = max(eps, tikhonov_reg)
    kernel_Xnys_sq = kernel_Xnys.T @ kernel_Xnys
    add_diagonal(kernel_X, reg * dim)
    if svd_solver == "full":
        values, vectors = eigh(
            kernel_Xnys_sq, kernel_X
        )  # normalization leads to needing to invert evals
    elif svd_solver == "arnoldi":
        _oversampling = max(10, 4 * int(np.sqrt(rank)))
        _num_arnoldi_eigs = min(rank + _oversampling, kernel_X.shape[0])
        values, vectors = eigsh(
            kernel_Xnys_sq,
            M=kernel_X,
            k=_num_arnoldi_eigs,
            which="LM",
        )
    else:
        raise ValueError(f"Unknown svd_solver {svd_solver}")
    add_diagonal(kernel_X, -reg * dim)
    vectors, values, rsqrt_values = eigh_rank_reveal(values, vectors, rank)

    U = np.sqrt(dim) * vectors * (rsqrt_values)
    V = np.linalg.multi_dot([kernel_Ynys.T, kernel_Xnys, vectors])
    V = lstsq(kernel_Y, V)[0]
    V = np.sqrt(dim) * V * (rsqrt_values)

    kernel_X_eigvalsh = (np.flip(np.sort(np.abs(values))) / kernel_Xnys.shape[0]) ** 0.5
    U, V = postprocess_UV(U, V, rank)
    return U, V, kernel_X_eigvalsh


def fit_rand_principal_component_regression(
    kernel_X: np.ndarray,  # Kernel matrix of the input data
    tikhonov_reg: float,  # Tikhonov regularization parameter
    rank: int,  # Rank of the estimator
    n_oversamples: int,  # Number of oversamples
    iterated_power: int,  # Number of iterations for the power method
    rng_seed: Optional[int] = None,  # Seed for the random number generator
):
    dim = kernel_X.shape[0]
    add_diagonal(kernel_X, dim * tikhonov_reg)
    vectors, values, _ = randomized_svd(
        kernel_X,
        rank,
        n_oversamples=n_oversamples,
        n_iter=iterated_power,
        random_state=rng_seed,
    )
    add_diagonal(kernel_X, -dim * tikhonov_reg)

    filtered_vectors, filtered_values, rsqrt_values = eigh_rank_reveal(
        values, vectors, rank
    )
    Q = np.sqrt(dim) * filtered_vectors * (rsqrt_values)
    kernel_X_eigvalsh = (np.flip(np.sort(np.abs(values))) / dim) ** 0.5
    Q, Q = postprocess_UV(Q, Q, rank)
    return Q, Q, kernel_X_eigvalsh


def predict(
    num_steps: int,  # Number of steps to predict (return the last one)
    U: np.ndarray,  # Projection matrix: first output of the fit functions defined above
    V: np.ndarray,  # Projection matrix: second output of the fit functions defined above
    K_YX: np.ndarray,  # Kernel matrix between the output data and the input data
    K_Xin_X: np.ndarray,  # Kernel matrix between the initial conditions and the input data
    obs_train_Y: np.ndarray,  # Observable to be predicted evaluated on the output training data
) -> np.ndarray:
    # G = S UV.T Z
    # G^n = (SU)(V.T K_YX U)^(n-1)(V.T Z)
    dim = U.shape[0]
    rsqrt_dim = dim ** (-0.5)
    K_dot_U = rsqrt_dim * K_Xin_X @ U
    V_dot_obs = rsqrt_dim * V.T @ obs_train_Y
    V_K_YX_U = (dim**-1) * np.linalg.multi_dot([V.T, K_YX, U])
    M = np.linalg.matrix_power(V_K_YX_U, num_steps - 1)
    return np.linalg.multi_dot([K_dot_U, M, V_dot_obs])


def estimator_eig(
    U: np.ndarray,  # Projection matrix: first output of the fit functions defined above
    V: np.ndarray,  # Projection matrix: second output of the fit functions defined above
    K_X: np.ndarray,  # Kernel matrix of the input data
    K_YX: np.ndarray,  # Kernel matrix between the output data and the input data
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    # SUV.TZ -> V.T K_YX U (right ev = SUvr, left ev = ZVvl)
    r_dim = (K_X.shape[0]) ** (-1)

    W_YX = np.linalg.multi_dot([V.T, r_dim * K_YX, U])
    W_X = np.linalg.multi_dot([U.T, r_dim * K_X, U])

    values, vl, vr = eig(W_YX, left=True, right=True)  # Left -> V, Right -> U
    values = fuzzy_parse_complex(values)
    r_perm = np.argsort(values)
    vr = vr[:, r_perm]
    l_perm = np.argsort(values.conj())
    vl = vl[:, l_perm]
    values = values[r_perm]

    rcond = 1000.0 * np.finfo(U.dtype).eps
    # Normalization in RKHS
    norm_r = weighted_norm(vr, W_X)
    norm_r = np.where(norm_r < rcond, np.inf, norm_r)
    vr = vr / norm_r

    # Bi-orthogonality of left eigenfunctions
    norm_l = np.diag(np.linalg.multi_dot([vl.T, W_YX, vr]))
    norm_l = np.where(np.abs(norm_l) < rcond, np.inf, norm_l)
    vl = vl / norm_l
    return values, V @ vl, U @ vr


def estimator_modes(K_Xin_X: np.ndarray, rv: np.ndarray, lv: np.ndarray):
    r_dim = lv.shape[0] ** -0.5
    rv_in = evaluate_eigenfunction(K_Xin_X, rv).T  # [rank, num_initial_conditions]
    lv_obs = r_dim * lv.T  # [rank, num_observations]
    return (
        rv_in[:, :, None] * lv_obs[:, None, :]
    )  # [rank, num_init_conditions, num_training_points]


def evaluate_eigenfunction(
    K_Xin_X_or_Y: np.ndarray,
    # Kernel matrix between the initial conditions and the input data (right eigenfunctions) or the output data
    # (left eigenfunctions)
    vr_or_vl: np.ndarray,  # Right eigenvectors or left eigenvectors, as returned by the estimator_eig function
):
    rsqrt_dim = (K_Xin_X_or_Y.shape[1]) ** (-0.5)
    return np.linalg.multi_dot([rsqrt_dim * K_Xin_X_or_Y, vr_or_vl])


def svdvals(
    U: np.ndarray,  # Projection matrix: first output of the fit functions defined above
    V: np.ndarray,  # Projection matrix: second output of the fit functions defined above
    K_X: np.ndarray,  # Kernel matrix of the input data
    K_Y: np.ndarray,  # Kernel matrix of the output data
):
    # Inefficient implementation
    rdim = (K_X.shape[0]) ** (-1)
    A = np.linalg.multi_dot([V.T, rdim * K_Y, V])
    B = np.linalg.multi_dot([U.T, rdim * K_X, U])
    v = eig(A @ B, left=False, right=False)
    # Clip the negative values
    v = v.real
    v[v < 0] = 0
    return np.sqrt(v)


def estimator_risk(
    kernel_Yv: np.ndarray,  # Kernel matrix of the output validation data
    kernel_Y: np.ndarray,  # Kernel matrix of the output training data
    kernel_XXv: np.ndarray,  # Cross-Kernel matrix of the input train/validation data
    kernel_YYv: np.ndarray,  # Cross-Kernel matrix of the output train/validation data
    U: np.ndarray,  # Projection matrix: first output of the fit functions defined above
    V: np.ndarray,  # Projection matrix: second output of the fit functions defined above
):
    rdim_train = (kernel_Y.shape[0]) ** (-1)
    rdim_val = (kernel_Yv.shape[0]) ** (-1)

    r_Y = rdim_val * np.trace(kernel_Yv)
    r_XY = (
        -2
        * rdim_val
        * rdim_train
        * np.trace(np.linalg.multi_dot([kernel_YYv.T, V, U.T, kernel_XXv]))
    )
    r_X = (
        rdim_val
        * (rdim_train**2)
        * np.trace(
            np.linalg.multi_dot([kernel_XXv.T, U, V.T, kernel_Y, V, U.T, kernel_XXv])
        )
    )
    return r_Y + r_XY + r_X



import numpy as np
import pytest

from kooplearn._src.operator_regression import dual, primal
from kooplearn.datasets import Mock


@pytest.mark.parametrize("scale_factor", [0.5, 2])
@pytest.mark.parametrize("tikhonov_reg", [0.0, 1e-3])
@pytest.mark.parametrize("svd_solver", ["full", "arnoldi"])
def test_reduced_rank_tikhonov_primal_scale_invariance(
    svd_solver, tikhonov_reg, scale_factor
):
    num_features = 10
    num_test_pts = 100
    rank = 5

    dataset = Mock(num_features=num_features, rng_seed=42)
    _Z = dataset.sample(None, num_test_pts)
    X, Y = _Z[:-1], _Z[1:]

    rdim = np.true_divide(1, X.shape[0])
    C_X = rdim * ((X.T) @ X)
    C_XY = rdim * ((X.T) @ Y)

    # Primal
    U = primal.fit_reduced_rank_regression(
        C_X, C_XY, tikhonov_reg, rank, svd_solver=svd_solver
    )
    U_scaled = primal.fit_reduced_rank_regression(
        scale_factor * C_X,
        scale_factor * C_XY,
        tikhonov_reg * scale_factor,
        rank,
        svd_solver=svd_solver,
    )

    G = np.linalg.multi_dot([U, U.T, C_XY])
    G_scaled = np.linalg.multi_dot([U_scaled, U_scaled.T, scale_factor * C_XY])
    assert np.allclose(G, G_scaled)


@pytest.mark.parametrize("scale_factor", [0.5, 2.0])
@pytest.mark.parametrize("tikhonov_reg", [1e-3])
@pytest.mark.parametrize("svd_solver", ["full", "arnoldi"])
def test_reduced_rank_tikhonov_dual_scale_invariance(
    svd_solver, tikhonov_reg, scale_factor
):
    num_features = 10
    num_test_pts = 100
    rank = 5

    dataset = Mock(num_features=num_features, rng_seed=42)
    _Z = dataset.sample(None, num_test_pts)
    X, Y = _Z[:-1], _Z[1:]

    K_X = X @ (X.T)
    K_Y = Y @ (Y.T)

    # Dual
    U, V, _ = dual.fit_reduced_rank_regression(
        K_X, K_Y, tikhonov_reg, rank, svd_solver=svd_solver
    )
    U_scaled, V_scaled, _ = dual.fit_reduced_rank_regression(
        scale_factor * K_X,
        scale_factor * K_Y,
        tikhonov_reg * scale_factor,
        rank,
        svd_solver=svd_solver,
    )

    G = np.linalg.multi_dot([U, V.T])
    G_scaled = np.linalg.multi_dot([U_scaled, V_scaled.T])
    assert np.allclose(G * (scale_factor**-1), G_scaled, atol=1e-5)



import logging
import weakref
from copy import deepcopy
from typing import Callable, Optional

import numpy as np
from scipy.linalg import eig

from kooplearn._src.check_deps import check_torch_deps
from kooplearn._src.serialization import pickle_load, pickle_save
from kooplearn._src.utils import ShapeError, check_contexts_shape, check_is_fitted
from kooplearn.abc import BaseModel
from kooplearn.models.ae.utils import (
    consistency_loss,
    decode_contexts,
    encode_contexts,
    evolve_batch,
    evolve_contexts,
)

logger = logging.getLogger("kooplearn")
check_torch_deps()
import lightning  # noqa: E402
import torch  # noqa: E402


class ConsistentAE(BaseModel):
    """Consistent AutoEncoder introduced by :footcite:t:`Azencot2020`.

    Args:
        encoder (torch.nn.Module): Encoder network. Will be initialized as ``encoder(**encoder_kwargs)``.
        decoder (torch.nn.Module): Decoder network. Will be initialized as ``decoder(**decoder_kwargs)``.
        latent_dim (int): Latent dimension. In must match the dimension of the outputs of ``encoder``, as well the dimensions of the inputs of ``decoder``.
        optimizer_fn (torch.optim.Optimizer): Any optimizer from :class:`torch.optim.Optimizer`.
        trainer (lightning.Trainer): An initialized `Lightning Trainer <https://lightning.ai/docs/pytorch/stable/common/trainer.html>`_ object used to train the Consistent AutoEncoder.
        loss_weights (dict, optional): Weights of the different loss terms. Should be a dictionary containing either some of all the keys ``rec`` (reconstruction loss), ``pred`` (prediction loss), ``bwd_pred`` (backward prediction loss), ``lin`` (linear evolution loss), and ``consistency`` (consistencu loss). Defaults to ``{ "rec": 1.0, "pred": 1.0, "bwd_pred": 1.0, "lin": 1.0, "consistency": 1.0}`` .
        encoder_kwargs (dict, optional): Dictionary of keyword arguments passed to the encoder network upon initialization. Defaults to ``{}``.
        decoder_kwargs (dict, optional): Dictionary of keyword arguments passed to the decoder network upon initialization. Defaults to ``{}``.
        optimizer_kwargs (dict, optional): Dictionary of keyword arguments passed to the optimizer at initialization. Defaults to ``{}``.
        backward_steps (int, optional): Number of steps of backward dynamics to perform. If ``backward_steps == 0``, this model reduces to :class:`kooplearn.models.DynamicAE`. Defaults to 1.
        seed (Optional[int], optional): Seed of the internal random number generator. Defaults to None.
    """

    def __init__(
        self,
        encoder: torch.nn.Module,
        decoder: torch.nn.Module,
        latent_dim: int,
        optimizer_fn: torch.optim.Optimizer,
        trainer: lightning.Trainer,
        loss_weights: dict = {
            "rec": 1.0,
            "pred": 1.0,
            "bwd_pred": 1.0,
            "lin": 1.0,
            "consistency": 1.0,
        },
        encoder_kwargs: dict = {},
        decoder_kwargs: dict = {},
        optimizer_kwargs: dict = {},
        backward_steps: int = 1,
        seed: Optional[int] = None,
    ):
        lightning.seed_everything(seed)
        self.lightning_trainer = trainer
        self.lightning_module = ConsistentAEModule(
            encoder,
            decoder,
            latent_dim,
            optimizer_fn,
            optimizer_kwargs,
            loss_weights=loss_weights,
            encoder_kwargs=encoder_kwargs,
            decoder_kwargs=decoder_kwargs,
            kooplearn_model_weakref=weakref.ref(self),
        )
        self.seed = seed
        self._is_fitted = False
        if backward_steps < 0:
            raise ValueError(
                f"backward_steps must be >= 0, got {backward_steps} instead."
            )
        else:
            self._bwd_steps = backward_steps
        # Todo: Add warning on lookback_len for this model

    def fit(
        self,
        train_dataloaders=None,
        val_dataloaders=None,
        datamodule: Optional[lightning.LightningDataModule] = None,
        ckpt_path: Optional[str] = None,
    ):
        """Fits the Koopman AutoEncoder model. Accepts the same arguments as :meth:`lightning.Trainer.fit`, except for the ``model`` keyword, which is automatically set internally.

        Args:
            train_dataloaders: An iterable or collection of iterables specifying training samples.
                Alternatively, a :class:`~lightning.pytorch.core.datamodule.LightningDataModule` that defines
                the :class:`~lightning.pytorch.core.hooks.DataHooks.train_dataloader` hook.

            val_dataloaders: An iterable or collection of iterables specifying validation samples.

            datamodule: A :class:`~lightning.pytorch.core.datamodule.LightningDataModule` that defines
                the :class:`~lightning.pytorch.core.hooks.DataHooks.train_dataloader` hook.

            ckpt_path: Path/URL of the checkpoint from which training is resumed. Could also be one of two special
                keywords ``"last"`` and ``"hpc"``. If there is no checkpoint file at the path, an exception is raised.
        """
        if isinstance(train_dataloaders, lightning.LightningDataModule):
            datamodule = train_dataloaders
            train_dataloaders = None

        # If you supply a datamodule you can't supply train_dataloader or val_dataloaders
        if (
            train_dataloaders is not None or val_dataloaders is not None
        ) and datamodule is not None:
            raise ValueError(
                "You cannot pass `train_dataloader` or `val_dataloaders` to `fit(datamodule=...)`"
            )
        # Get the shape of the first batch to determine the lookback_len
        if train_dataloaders is None:
            assert isinstance(datamodule, lightning.LightningDataModule)
            for batch in datamodule.train_dataloader():
                with torch.no_grad():
                    self.lightning_module.dry_run(batch)
                    self._state_trail_dims = tuple(batch.shape[2:])
                break
        else:
            assert isinstance(train_dataloaders, torch.utils.data.DataLoader)
            for batch in train_dataloaders:
                with torch.no_grad():
                    self.lightning_module.dry_run(batch)
                    self._state_trail_dims = tuple(batch.shape[2:])
                break

        self.lightning_trainer.fit(
            model=self.lightning_module,
            train_dataloaders=train_dataloaders,
            val_dataloaders=val_dataloaders,
            datamodule=datamodule,
            ckpt_path=ckpt_path,
        )
        self._is_fitted = True

    def _np_to_torch(self, data: np.ndarray):
        check_contexts_shape(data, self.lookback_len, is_inference_data=True)
        model_device = self.lightning_module.device
        return torch.from_numpy(data.copy()).float().to(model_device)

    def predict(
        self,
        data: np.ndarray,
        t: int = 1,
        observables: Optional[Callable] = None,
    ):
        """
        Predicts the state or, if the system is stochastic, its expected value :math:`\mathbb{E}[X_t | X_0 = X]` after ``t`` instants given the initial conditions ``X = data[:, self.lookback_len:, ...]`` being the lookback slice of ``data``.

        .. attention::

            ``data.shape[1]`` must match the lookback length ``self.lookback_len``. Otherwise, an error is raised.

        If ``observables`` are not ``None``, returns the analogue quantity for the observable instead.

        Args:
            data (numpy.ndarray): Initial conditions to predict. Array of context windows with shape ``(n_init_conditions, self.lookback_len, *self.data_fit.shape[2:])`` (see the note above).
            t (int): Number of steps in the future to predict (returns the last one).
            observables (callable or None): Callable or ``None``. If callable should map batches of states of shape ``(batch, *self.data_fit.shape[2:])`` to batches of observables ``(batch, *obs_features_shape)``.

        Returns:
           The predicted (expected) state/observable at time :math:`t`, shape ``(n_init_conditions, *obs_features_shape)``.
        """
        torch_data = self._np_to_torch(data)  # [n_samples, context_len, *trail_dims]
        check_is_fitted(self, ["_state_trail_dims"])
        assert tuple(torch_data.shape[2:]) == self._state_trail_dims

        with torch.no_grad():
            init_data = torch_data[:, self.lookback_len - 1, ...]
            evolved_data = evolve_batch(
                init_data,
                t,
                self.lightning_module.encoder,
                self.lightning_module.decoder,
                self.lightning_module.evolution_operator,
            )
            evolved_data = evolved_data.detach().cpu().numpy()
        if observables is None:
            return evolved_data
        elif callable(observables):
            return observables(evolved_data)
        else:
            raise ValueError("Observables must be either None, or callable.")

    def modes(
        self,
        data: np.ndarray,
        observables: Optional[Callable] = None,
    ):
        raise NotImplementedError()

    def eig(
        self,
        eval_left_on: Optional[np.ndarray] = None,
        eval_right_on: Optional[np.ndarray] = None,
    ):
        if hasattr(self, "_eig_cache"):
            w, vl, vr = self._eig_cache
        else:
            K = self.lightning_module.evolution_operator
            K_np = K.detach().cpu().numpy()
            w, vl, vr = eig(K_np, left=True, right=True)
            self._eig_cache = w, vl, vr

        if eval_left_on is None and eval_right_on is None:
            # (eigenvalues,)
            return w
        elif eval_left_on is None and eval_right_on is not None:
            # (eigenvalues, right eigenfunctions)
            eval_right_on = self._np_to_torch(
                eval_right_on
            )  # [n_samples, context_len == 1, *trail_dims]
            eval_right_on = eval_right_on[
                :, self.lookback_len - 1, ...
            ]  # [n_samples, *trail_dims]
            with torch.no_grad():
                phi_Xin = self.lightning_module.encoder(eval_right_on)
                r_fns = (
                    phi_Xin @ vl
                )  # Not a typo: I need the left eigenvectors of K to get the right eigenfunctions of the Koopman operator
            return w, r_fns.detach().cpu().numpy()
        elif eval_left_on is not None and eval_right_on is None:
            # (eigenvalues, left eigenfunctions)
            eval_left_on = self._np_to_torch(
                eval_left_on
            )  # [n_samples, context_len == 1, *trail_dims]
            eval_left_on = eval_left_on[
                :, self.lookback_len - 1, ...
            ]  # [n_samples, *trail_dims]
            with torch.no_grad():
                phi_Xin = self.lightning_module.encoder(eval_left_on)
                l_fns = (
                    phi_Xin @ vr
                )  # Not a typo: I need the right eigenvectors of K to get the left eigenfunctions of the Koopman operator
            return w, l_fns.detach().cpu().numpy()
        elif eval_left_on is not None and eval_right_on is not None:
            # (eigenvalues, left eigenfunctions, right eigenfunctions)

            eval_right_on = self._np_to_torch(
                eval_right_on
            )  # [n_samples, context_len == 1, *trail_dims]
            eval_right_on = eval_right_on[
                :, self.lookback_len - 1, ...
            ]  # [n_samples, *trail_dims]

            eval_left_on = self._np_to_torch(
                eval_left_on
            )  # [n_samples, context_len == 1, *trail_dims]
            eval_left_on = eval_left_on[
                :, self.lookback_len - 1, ...
            ]  # [n_samples, *trail_dims]

            with torch.no_grad():
                phi_Xin_r = self.lightning_module.encoder(eval_right_on)
                r_fns = (
                    phi_Xin_r @ vl
                )  # Not a typo: I need the left eigenvectors of K to get the right eigenfunctions of the Koopman operator

                phi_Xin_l = self.lightning_module.encoder(eval_left_on)
                l_fns = (
                    phi_Xin_l @ vr
                )  # Not a typo: I need the right eigenvectors of K to get the left eigenfunctions of the Koopman operator

            return w, l_fns.detach().cpu().numpy(), r_fns.detach().cpu().numpy()

    def save(self, filename):
        """Serialize the model to a file.

        Args:
            filename (path-like or file-like): Save the model to file.
        """
        # Delete (un-picklable) weakref self.lightning_module._kooplearn_feature_map_weakref
        self.lightning_module._kooplearn_model_weakref = None
        pickle_save(self, filename)

    @classmethod
    def load(cls, filename):
        """Load a serialized model from a file.

        Args:
            filename (path-like or file-like): Load the model from file.

        Returns:
            ConsistentAE: The loaded model.
        """
        restored_obj = pickle_load(cls, filename)
        # Restore the weakref
        restored_obj.lightning_module._kooplearn_model_weakref = weakref.ref(
            restored_obj
        )
        return restored_obj

    @property
    def is_fitted(self) -> bool:
        return self._is_fitted

    @property
    def lookback_len(self) -> int:
        return 1 + self._bwd_steps


class ConsistentAEModule(lightning.LightningModule):
    def __init__(
        self,
        encoder: torch.nn.Module,
        decoder: torch.nn.Module,
        latent_dim: int,
        optimizer_fn: torch.optim.Optimizer,
        optimizer_kwargs: dict,
        loss_weights: dict = {
            "rec": 1.0,
            "pred": 1.0,
            "bwd_pred": 1.0,
            "lin": 1.0,
            "consistency": 1.0,
        },
        encoder_kwargs: dict = {},
        decoder_kwargs: dict = {},
        kooplearn_model_weakref: weakref.ReferenceType = None,
    ):
        super().__init__()
        self.save_hyperparameters(ignore=["kooplearn_model_weakref", "optimizer_fn"])
        self.encoder = encoder(**encoder_kwargs)
        self.decoder = decoder(**decoder_kwargs)
        self._lin = torch.nn.Linear(latent_dim, latent_dim, bias=False)
        self._bwd_lin = torch.nn.Linear(latent_dim, latent_dim, bias=False)
        self.evolution_operator = self._lin.weight
        self.bwd_evolution_operator = self._bwd_lin.weight

        self._optimizer = optimizer_fn
        _tmp_opt_kwargs = deepcopy(optimizer_kwargs)
        if "lr" in _tmp_opt_kwargs:  # For Lightning's LearningRateFinder
            self.lr = _tmp_opt_kwargs.pop("lr")
            self.opt_kwargs = _tmp_opt_kwargs
        else:
            self.lr = 1e-3
            logger.warning(
                "No learning rate specified. Using default value of 1e-3. You can specify the learning rate by passing it to the optimizer_kwargs argument."
            )
        self._kooplearn_model_weakref = kooplearn_model_weakref

    def configure_optimizers(self):
        kw = self.opt_kwargs | {"lr": self.lr}
        return self._optimizer(self.parameters(), **kw)

    def training_step(self, train_batch, batch_idx):
        lookback_len = self._kooplearn_model_weakref().lookback_len
        encoded_batch = encode_contexts(train_batch, self.encoder)

        K = self.evolution_operator
        bwd_K = self.bwd_evolution_operator

        evolved_batch = evolve_contexts(
            encoded_batch, lookback_len, K, backward_operator=bwd_K
        )
        decoded_batch = decode_contexts(evolved_batch, self.decoder)

        MSE = torch.nn.MSELoss()
        # Reconstruction + prediction loss
        rec_loss = MSE(
            train_batch[:, lookback_len - 1, ...],
            decoded_batch[:, lookback_len - 1, ...],
        )

        pred_loss = MSE(
            train_batch[:, lookback_len:, ...], decoded_batch[:, lookback_len:, ...]
        )
        bwd_pred_loss = MSE(
            train_batch[:, : (lookback_len - 1), ...],
            decoded_batch[:, : (lookback_len - 1), ...],
        )
        # Linear loss
        lin_loss = MSE(encoded_batch, evolved_batch)
        # Consistency loss
        cnst_loss = consistency_loss(
            self.evolution_operator, self.bwd_evolution_operator
        )

        alpha_rec = self.hparams.loss_weights.get("rec", 1.0)
        alpha_pred = self.hparams.loss_weights.get("pred", 1.0)
        alpha_bwd_pred = self.hparams.loss_weights.get("bwd_pred", 1.0)
        alpha_lin = self.hparams.loss_weights.get("lin", 1.0)
        alpha_consistency = self.hparams.loss_weights.get("consistency", 1.0)

        loss = (
            alpha_rec * rec_loss
            + alpha_pred * pred_loss
            + alpha_bwd_pred * bwd_pred_loss
            + alpha_lin * lin_loss
            + alpha_consistency * cnst_loss
        )
        metrics = {
            "train/reconstruction_loss": rec_loss.item(),
            "train/prediction_loss": pred_loss.item(),
            "train/backward_prediction_loss": bwd_pred_loss.item(),
            "train/linear_loss": lin_loss.item(),
            # "train/consistency_loss": cnst_loss.item(),
            "train/full_loss": loss.item(),
        }
        self.log_dict(metrics, on_step=True, prog_bar=True, logger=True)
        return loss

    def dry_run(self, batch: torch.Tensor):
        lookback_len = self._kooplearn_model_weakref().lookback_len
        check_contexts_shape(batch, lookback_len)
        # Caution: this method is designed only for internal calling.
        Z = encode_contexts(batch, self.encoder)

        evolution_operator = self.evolution_operator
        bwd_evolution_operator = self.bwd_evolution_operator

        Z_evolved = evolve_contexts(
            Z,
            lookback_len,
            evolution_operator,
            backward_operator=bwd_evolution_operator,
        )
        X_evol = decode_contexts(
            Z_evolved, self.decoder
        )  # Should fail if the shape is wrong
        assert Z.shape == Z_evolved.shape

        if batch.shape != X_evol.shape:
            raise ShapeError(
                f"The shape of the evolved states {X_evol.shape[2:]} does not match the shape of the initial states {batch.shape[2:]}. Please check that the decoder networks output tensors of the same shape as the input tensors."
            )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

sys.path.insert(0, os.path.abspath(".."))  # Importing kooplearn for autodoc

project = "kooplearn"
copyright = "2023, Pietro Novelli"
author = "Pietro Novelli"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinxcontrib.bibtex",
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_design",
    "myst_nb",
]

source_suffix = {
    ".rst": "restructuredtext",
    ".ipynb": "myst-nb",
    ".myst": "myst-nb",
}

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
autodoc_typehints = "none"
autodoc_class_signature = "separated"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output
html_theme = "sphinx_book_theme"
html_logo = "../logo.svg"

html_theme_options = {
    "repository_url": "https://github.com/Machine-Learning-Dynamical-Systems/kooplearn",
    "use_repository_button": True,  # add a 'link to repository' button
    "use_issues_button": False,  # add an 'Open an Issue' button
    "path_to_docs": ("docs"),
    "show_navbar_depth": 1,
    "show_toc_level": 3,
}

bibtex_bibfiles = ["bibliography.bib"]

myst_enable_extensions = ["amsmath", "dollarmath", "html_image"]



import numpy as np

from kooplearn._src.utils import ShapeError, check_contexts_shape


def parse_observables(observables, data, data_fit, lookback_len):
    # Shape checks:
    check_contexts_shape(data, lookback_len, is_inference_data=True)
    check_contexts_shape(data_fit, lookback_len)
    data = np.asanyarray(data)

    X_inference, _ = contexts_to_markov_predict_states(data, lookback_len)
    X_fit, Y_fit = contexts_to_markov_predict_states(data_fit, lookback_len)

    if observables is None:
        _obs = Y_fit
    elif callable(observables):
        _obs = observables(Y_fit)
    else:
        raise ValueError("Observables must be either None, or callable.")

    # Reshape the observables to 2D arrays
    if _obs.ndim == 1:
        _obs = _obs[:, None]

    # If the observables are multidimensional, flatten them and save the shape for the final reshape
    _obs_trailing_dims = _obs.shape[1:]
    expected_shape = (X_inference.shape[0],) + _obs_trailing_dims
    if _obs.ndim > 2:
        _obs = _obs.reshape(_obs.shape[0], -1)
    return _obs, expected_shape, X_inference, X_fit


def contexts_to_markov_train_states(
    contexts: np.ndarray,
    lookback_len: int,
) -> np.ndarray:
    """TODO: Docstring for contexts_to_markov_IO_states.

    Args:
        contexts (np.ndarray): Array of contexts with shape ``(n_samples, context_len, *features_shape)``
        lookback_len (int): Length of the lookback window associated to the contexts.

        .. caution::

        If the lookforward window ``context_len - lookback_len - 1`` is not equal to 1, an error will be raised.

    Returns:
        tuple(np.ndarray, np.ndarray): TODO.
    """
    if not (contexts.shape[1] == lookback_len + 1):
        raise ShapeError(
            f"This function act on context windows with lookforward dimension == 1, while context_len = {contexts.shape[1]} and lookback_len = {lookback_len}, that is lookforward_len = {contexts.shape[1] - lookback_len} != 1"
        )

    _init = contexts[:, :-1, ...]
    _evolution = contexts[:, 1:, ...]
    return _init, _evolution


def contexts_to_markov_predict_states(
    contexts: np.ndarray,
    lookback_len: int,
) -> np.ndarray:
    """TODO: Docstring for contexts_to_markov_IO_states.

    Args:
        contexts (np.ndarray): Array of contexts with shape ``(n_samples, context_len, *features_shape)``
        lookback_len (int): Length of the lookback window associated to the contexts.

        .. caution::

        If the lookforward window ``context_len - lookback_len - 1`` is larger than 1, an error will be raised.

    Returns:
        tuple(np.ndarray, np.ndarray): TODO.
    """

    if lookback_len == contexts.shape[1]:
        X = contexts
        Y = None
    elif lookback_len + 1 == contexts.shape[1]:
        X = contexts[:, :-1, ...]
        Y = contexts[:, -1, ...]
    else:
        raise ShapeError(
            f"This function act on context windows with lookforward dimension == 0 or 1, while context_len = {contexts.shape[1]} and lookback_len = {lookback_len}, that is lookforward_len = {contexts.shape[1] - lookback_len} != 0 or 1"
        )
    return X, Y



import logging
import weakref
from copy import deepcopy
from typing import Optional

import lightning
import numpy as np
import torch

from kooplearn._src.serialization import pickle_load, pickle_save
from kooplearn.abc import TrainableFeatureMap
from kooplearn.nn.functional import (
    log_fro_metric_deformation_loss,
    relaxed_projection_score,
    vamp_score,
)

logger = logging.getLogger("kooplearn")


class DPNet(TrainableFeatureMap):
    """Implements the DPNets :footcite:p:`Kostic2023DPNets` feature map, which learn an invariant representation of time-homogeneous stochastic dynamical systems. Can be used in conjunction to :class:`kooplearn.models.DeepEDMD` to learn a Koopman/Transfer operator from data. The DPNet feature map is trained using the :class:`lightning.LightningModule` API, and can be trained using the :class:`lightning.Trainer` API. See the `PyTorch Lightning documentation <https://pytorch-lightning.readthedocs.io/en/latest/>`_ for more information.

    Args:
        encoder (torch.nn.Module): Encoder network. Should be a subclass of :class:`torch.nn.Module`. Will be initialized as ``encoder(**encoder_kwargs)``.
        optimizer_fn (torch.optim.Optimizer): Any optimizer from :class:`torch.optim.Optimizer`.
        trainer (lightning.Trainer): An initialized `Lightning Trainer <https://lightning.ai/docs/pytorch/stable/common/trainer.html>`_ object used to train the DPNet feature map.
        use_relaxed_loss (bool, optional): Whether to use the relaxed projection score introduced in :footcite:t:`Kostic2023DPNets`. Might be slower to convergence, but is much more stable in ill-conditioned problems.  Defaults to False.
        metric_deformation_loss_coefficient (float, optional): Coefficient of the metric deformation loss. Defaults to 1.0.
        encoder_kwargs (dict, optional): Dictionary of keyword arguments passed to the encoder network upon initialization. Defaults to ``{}``.
        optimizer_kwargs (dict): Dictionary of keyword arguments passed to the optimizer at initialization. Defaults to ``{}``.
        encoder_timelagged (Optional[torch.nn.Module], optional): Encoder network for the time-lagged data. Defaults to None. If None, the encoder network is used for time-lagged data as well. If not None, it will be initialized as ``encoder_timelagged(**encoder_timelagged_kwargs)``.
        encoder_timelagged_kwargs (dict, optional): Dictionary of keyword arguments passed to `encoder_timelagged` upon initialization. Defaults to ``{}``.
        center_covariances (bool, optional): Wheter to compute the VAMP score with centered covariances. Defaults to False.
        seed (int, optional): Seed of the internal random number generator. Defaults to None.
    """

    def __init__(
        self,
        encoder: torch.nn.Module,
        optimizer_fn: torch.optim.Optimizer,
        trainer: lightning.Trainer,
        use_relaxed_loss: bool = False,
        metric_deformation_loss_coefficient: float = 1.0,  # That is, the parameter γ in the paper.
        encoder_kwargs: dict = {},
        optimizer_kwargs: dict = {},
        encoder_timelagged: Optional[torch.nn.Module] = None,
        encoder_timelagged_kwargs: dict = {},
        center_covariances: bool = False,
        seed: Optional[int] = None,
    ):
        if seed is not None:
            lightning.seed_everything(seed)

        self.lightning_trainer = trainer
        self.lightning_module = DPModule(
            encoder,
            optimizer_fn,
            optimizer_kwargs,
            use_relaxed_loss=use_relaxed_loss,
            metric_deformation_loss_coefficient=metric_deformation_loss_coefficient,
            encoder_kwargs=encoder_kwargs,
            encoder_timelagged=encoder_timelagged,
            encoder_timelagged_kwargs=encoder_timelagged_kwargs,
            center_covariances=center_covariances,
            kooplearn_feature_map_weakref=weakref.ref(self),
        )
        self.seed = seed
        self._lookback_len = -1  # Dummy init value, will be determined at fit time.
        self._is_fitted = False

    @property
    def is_fitted(self):
        return self._is_fitted

    @property
    def lookback_len(self):
        return self._lookback_len

    def save(self, filename):
        """Serialize the model to a file.

        Args:
            filename (path-like or file-like): Save the model to file.
        """
        # Delete (un-picklable) weakref self.lightning_module._kooplearn_feature_map_weakref
        self.lightning_module._kooplearn_feature_map_weakref = None
        pickle_save(self, filename)

    @classmethod
    def load(cls, filename):
        """Load a serialized model from a file.

        Args:
            filename (path-like or file-like): Load the model from file.

        Returns:
            DPNet: The loaded model.
        """
        restored_obj = pickle_load(cls, filename)
        # Restore the weakref
        restored_obj.lightning_module._kooplearn_feature_map_weakref = weakref.ref(
            restored_obj
        )
        return restored_obj

    def fit(
        self,
        train_dataloaders=None,
        val_dataloaders=None,
        datamodule: Optional[lightning.LightningDataModule] = None,
        ckpt_path: Optional[str] = None,
        verbose: bool = True,
    ):
        """Fits the DPNet feature map. Accepts the same arguments as :meth:`lightning.Trainer.fit`, except for the ``model`` keyword, which is automatically set to the DPNet feature map.

        Args:
            train_dataloaders: An iterable or collection of iterables specifying training samples.
                Alternatively, a :class:`~lightning.pytorch.core.datamodule.LightningDataModule` that defines
                the :class:`~lightning.pytorch.core.hooks.DataHooks.train_dataloader` hook.

            val_dataloaders: An iterable or collection of iterables specifying validation samples.

            datamodule: A :class:`~lightning.pytorch.core.datamodule.LightningDataModule` that defines
                the :class:`~lightning.pytorch.core.hooks.DataHooks.train_dataloader` hook.

            ckpt_path: Path/URL of the checkpoint from which training is resumed. Could also be one of two special
                keywords ``"last"`` and ``"hpc"``. If there is no checkpoint file at the path, an exception is raised.
        """
        if isinstance(train_dataloaders, lightning.LightningDataModule):
            datamodule = train_dataloaders
            train_dataloaders = None

        # If you supply a datamodule you can't supply train_dataloader or val_dataloaders
        if (
            train_dataloaders is not None or val_dataloaders is not None
        ) and datamodule is not None:
            raise ValueError(
                "You cannot pass `train_dataloader` or `val_dataloaders` to `DPNet.fit(datamodule=...)`"
            )
        # Get the shape of the first batch to determine the lookback_len
        if train_dataloaders is None:
            assert isinstance(datamodule, lightning.LightningDataModule)
            for batch in datamodule.train_dataloader():
                context_len = batch.shape[1]
                break
        else:
            assert isinstance(train_dataloaders, torch.utils.data.DataLoader)
            for batch in train_dataloaders:
                context_len = batch.shape[1]
                break

        self._lookback_len = context_len - 1
        if verbose:
            print(
                f"Fitting {self.__class__.__name__}. Lookback window length set to {self.lookback_len}"
            )
        self.lightning_trainer.fit(
            model=self.lightning_module,
            train_dataloaders=train_dataloaders,
            val_dataloaders=val_dataloaders,
            datamodule=datamodule,
            ckpt_path=ckpt_path,
        )
        self._is_fitted = True

    def __call__(self, X: np.ndarray) -> np.ndarray:
        X = torch.from_numpy(X.copy(order="C")).float()
        self.lightning_module.eval()
        with torch.no_grad():
            embedded_X = self.lightning_module.encoder(
                X.to(self.lightning_module.device)
            )
            embedded_X = embedded_X.detach().cpu().numpy()
        return embedded_X


class DPModule(lightning.LightningModule):
    def __init__(
        self,
        encoder: torch.nn.Module,
        optimizer_fn: torch.optim.Optimizer,
        optimizer_kwargs: dict,
        use_relaxed_loss: bool = False,
        metric_deformation_loss_coefficient: float = 1.0,  # That is, the parameter γ in the paper.
        encoder_kwargs: dict = {},
        encoder_timelagged: Optional[torch.nn.Module] = None,
        encoder_timelagged_kwargs: dict = {},
        center_covariances: bool = True,
        kooplearn_feature_map_weakref=None,
    ):
        super().__init__()

        self.save_hyperparameters(
            ignore=["encoder", "optimizer_fn", "kooplearn_feature_map_weakref"]
        )
        _tmp_opt_kwargs = deepcopy(optimizer_kwargs)
        if "lr" in _tmp_opt_kwargs:  # For Lightning's LearningRateFinder
            self.lr = _tmp_opt_kwargs.pop("lr")
            self.opt_kwargs = _tmp_opt_kwargs
        else:
            self.lr = 1e-3
            logger.warning(
                "No learning rate specified. Using default value of 1e-3. You can specify the learning rate by passing it to the optimizer_kwargs argument."
            )

        self.encoder = encoder(**encoder_kwargs)
        if encoder_timelagged is not None:
            self.encoder_timelagged = encoder_timelagged(**encoder_timelagged_kwargs)
        else:
            self.encoder_timelagged = self.encoder
        self._optimizer = optimizer_fn
        self._kooplearn_feature_map_weakref = kooplearn_feature_map_weakref

    def configure_optimizers(self):
        kw = self.opt_kwargs | {"lr": self.lr}
        return self._optimizer(self.parameters(), **kw)

    def training_step(self, train_batch, batch_idx):
        X, Y = train_batch[:, :-1, ...], train_batch[:, 1:, ...]
        encoded_X, encoded_Y = self.forward(X), self.forward(Y, time_lagged=True)

        if self.hparams.center_covariances:
            encoded_X = encoded_X - encoded_X.mean(dim=0, keepdim=True)
            encoded_Y = encoded_Y - encoded_Y.mean(dim=0, keepdim=True)

        _norm = torch.rsqrt(torch.tensor(encoded_X.shape[0]))
        encoded_X *= _norm
        encoded_Y *= _norm

        cov_X = torch.mm(encoded_X.T, encoded_X)
        cov_Y = torch.mm(encoded_Y.T, encoded_Y)
        cov_XY = torch.mm(encoded_X.T, encoded_Y)

        metrics = {}
        # Compute the losses
        if self.hparams.use_relaxed_loss:
            svd_loss = -1 * relaxed_projection_score(cov_X, cov_Y, cov_XY)
            metrics["train/relaxed_projection_score"] = -1.0 * svd_loss.item()
        else:
            svd_loss = -1 * vamp_score(cov_X, cov_Y, cov_XY, schatten_norm=2)
            metrics["train/projection_score"] = -1.0 * svd_loss.item()
        if self.hparams.metric_deformation_loss_coefficient > 0.0:
            metric_deformation_loss = 0.5 * (
                log_fro_metric_deformation_loss(cov_X)
                + log_fro_metric_deformation_loss(cov_Y)
            )
            metric_deformation_loss *= self.hparams.metric_deformation_loss_coefficient
            metrics["train/metric_deformation_loss"] = metric_deformation_loss.item()
            svd_loss += metric_deformation_loss
        metrics["train/total_loss"] = svd_loss.item()
        self.log_dict(metrics, on_step=True, prog_bar=True, logger=True)
        return svd_loss

    def forward(self, X: torch.Tensor, time_lagged: bool = False) -> torch.Tensor:
        # Caution: this method is designed only for internal calling by the DPNet feature map.
        lookback_len = X.shape[1]
        batch_size = X.shape[0]
        trail_dims = X.shape[2:]
        X = X.view(lookback_len * batch_size, *trail_dims)
        if time_lagged:
            encoded_X = self.encoder_timelagged(X)
        else:
            encoded_X = self.encoder(X)
        trail_dims = encoded_X.shape[1:]
        encoded_X = encoded_X.view(batch_size, lookback_len, *trail_dims)
        return encoded_X.view(batch_size, -1)



from kooplearn.models.edmd import DMD, ExtendedDMD
from kooplearn.models.kernel import KernelDMD
from kooplearn.models.nystroem import NystroemKernel

from kooplearn.models.deepedmd import DeepEDMD  # isort:skip

try:
    from kooplearn._src.check_deps import check_torch_deps

    check_torch_deps()
    from kooplearn.models.ae.consistent import ConsistentAE
    from kooplearn.models.ae.dynamic import DynamicAE
except:
    pass



import pickle
from pathlib import Path


def pickle_save(obj, filename, protocol=5):
    # Check if 'filename' is a path-like or a file-like object
    if hasattr(filename, "write"):
        # If it is a file-like object, we dump the object to it
        pickle.dump(obj, filename, protocol=protocol)
    else:
        # If it is a path-like object, we dump the object to a file, creating the folder structure, if it does not exist
        filename = Path(filename)
        filename.parent.mkdir(exist_ok=True, parents=True)
        with open(filename, "+wb") as outfile:
            pickle.dump(obj, outfile, protocol=protocol)


def pickle_load(base_cls, filename):
    # Check if 'filename' is a path-like or a file-like object
    if hasattr(filename, "read"):
        # If it is a file-like object, we load the object from it
        restored_obj = pickle.load(filename)
    else:
        # If it is a path-like object, we load the object from a file
        with open(filename, "+rb") as infile:
            restored_obj = pickle.load(infile)
    assert type(restored_obj) == base_cls
    return restored_obj



import numpy as np
import pytest

from kooplearn._src.utils import parse_cplx_eig, topk

rng = np.random.default_rng(42)  # Global rng


@pytest.mark.parametrize("k", [0, 1, 3])
def test_topk(k):
    x = np.arange(10)
    if k == 0:
        with pytest.raises(AssertionError):
            res = topk(x, k=k)
    else:
        res = topk(x, k=k)
        assert np.allclose(res.values, np.flip(x)[:k])
        assert np.allclose(x[res.indices], np.flip(x)[:k])

    with pytest.raises(AssertionError):
        res = topk(rng.random((10, 20)), k=10)


@pytest.mark.parametrize("num_reals", [0, 1, 3])
@pytest.mark.parametrize("num_cplx_pairs", [0, 1, 3])
def test_parse_cplx_eig_(num_reals, num_cplx_pairs):
    rand_real = np.sort(rng.random(num_reals))

    rand_angle = rng.random(num_cplx_pairs) * np.pi
    rand_abs = rng.random(num_cplx_pairs)
    rand_cplx = np.sort(rand_abs * np.exp(1j * rand_angle))

    vec = np.concatenate([rand_real, rand_cplx, rand_cplx.conj()])

    rand_perm = rng.permutation(vec.shape[0])
    assert np.allclose(
        parse_cplx_eig(vec[rand_perm]), vec[: (num_reals + num_cplx_pairs)]
    )



from kooplearn.models.feature_maps.base import (
    ConcatenateFeatureMaps,
    IdentityFeatureMap,
)

try:
    from kooplearn._src.check_deps import check_torch_deps

    check_torch_deps()
    from kooplearn.models.feature_maps.dpnets import DPNet
    from kooplearn.models.feature_maps.vampnets import VAMPNet
except ImportError:
    pass



import shutil
from pathlib import Path

import ml_confs as mlcfg
from datasets import DatasetDict, interleave_datasets, load_dataset

main_path = Path(__file__).parent
data_path = main_path / "__data__"
configs = mlcfg.from_file(main_path / "configs.yaml")


def make_dataset():
    # Data pipeline
    MNIST = load_dataset("mnist", keep_in_memory=True)
    digit_ds = []
    for i in range(configs.classes):
        digit_ds.append(
            MNIST.filter(
                lambda example: example["label"] == i, keep_in_memory=True, num_proc=8
            )
        )
    ordered_MNIST = DatasetDict()
    # Order the digits in the dataset and select only a subset of the data
    for split in ["train", "test"]:
        ordered_MNIST[split] = interleave_datasets(
            [ds[split] for ds in digit_ds], split=split
        ).select(range(configs[f"{split}_samples"]))
    _tmp_ds = ordered_MNIST["train"].train_test_split(
        test_size=configs.val_ratio, shuffle=False
    )
    ordered_MNIST["train"] = _tmp_ds["train"]
    ordered_MNIST["validation"] = _tmp_ds["test"]
    ordered_MNIST.set_format(type="torch", columns=["image", "label"])
    ordered_MNIST = ordered_MNIST.map(
        lambda example: {"image": example["image"] / 255.0, "label": example["label"]},
        batched=True,
        keep_in_memory=True,
        num_proc=2,
    )
    ordered_MNIST.save_to_disk(data_path)
    configs.to_file(data_path / "configs.yaml")


if __name__ == "__main__":
    # Check if data_path exists, if not preprocess the data
    if not data_path.exists():
        print("Data directory not found, preprocessing data.")
        make_dataset()
    else:
        # Try to load the configs.yaml file and compare with the current one, if different, wipe the data_path and preprocess the data
        _saved_configs = mlcfg.from_file(data_path / "configs.yaml")
        if _saved_configs != configs:
            print("Configs changed, preprocessing data.")
            # Delete the data_path and preprocess the data
            shutil.rmtree(data_path)
            make_dataset()
        else:
            print("Data already preprocessed.")



import logging

import numpy as np

logger = logging.getLogger("kooplearn")

try:
    from kooplearn._src.check_deps import check_torch_deps

    check_torch_deps()
    from kooplearn.nn.data import ContextsDataset, traj_to_contexts_dataset
except:
    pass


def traj_to_contexts(
    trajectory: np.ndarray, context_window_len: int = 2, time_lag: int = 1
):
    """Convert a single trajectory to a sequence of context windows.

    Args:
        trajectory (np.ndarray): A trajectory of shape ``(n_frames, *features_shape)``.
        context_window_len (int, optional): Length of the context window. Defaults to 2.
        time_lag (int, optional): Time lag, i.e. stride, between successive context windows. Defaults to 1.

    Returns:
        np.ndarray: A sequence of context windows of shape ``(n_contexts, context_window_len, *features_shape)``.
    """
    if context_window_len < 2:
        raise ValueError(f"context_window_len must be >= 2, got {context_window_len}")

    if time_lag < 1:
        raise ValueError(f"time_lag must be >= 1, got {time_lag}")

    trajectory = np.asanyarray(trajectory)
    if trajectory.ndim == 0:
        trajectory = trajectory.reshape(1, 1)
    elif trajectory.ndim == 1:
        trajectory = trajectory[:, np.newaxis]

    _context_window_len = 1 + (context_window_len - 1) * time_lag
    if _context_window_len > trajectory.shape[0]:
        raise ValueError(
            f"Invalid combination of context_window_len={context_window_len} and time_lag={time_lag} for trajectory of length {trajectory.shape[0]}. Try reducing context_window_len or time_lag."
        )

    _res = np.lib.stride_tricks.sliding_window_view(
        trajectory, _context_window_len, axis=0
    )
    _res = np.moveaxis(_res, -1, 1)[:, ::time_lag, ...]
    return _res



import logging
import math
from typing import NamedTuple

import numpy as np
from numpy.typing import ArrayLike
from scipy.spatial.distance import pdist

logger = logging.getLogger("kooplearn")


# Exceptions
class NotFittedError(Exception):
    pass


class ShapeError(Exception):
    pass


# Misc Utils
def check_is_fitted(obj: object, attr_list: list[str]):
    for attr in attr_list:
        if not hasattr(obj, attr):
            raise NotFittedError(
                f"Attribute \"{attr}\" not found. {obj.__class__.__name__} is not fitted. Please call the 'fit' method first."
            )


def check_contexts_shape(
    data: ArrayLike, lookback_len: int, is_inference_data: bool = False
):
    # Numpy/Torch/Jax compatible
    if not isinstance(lookback_len, int):
        raise ValueError(
            f"The lookback_len must be an int, while {type(lookback_len)=}"
        )
    if lookback_len < 1:
        raise ValueError(f"Invalid lookback_len={lookback_len}.")
    if data.ndim < 3:
        raise ShapeError(
            f"Invalid shape {data.shape}. The data must have be at least three dimensional [batch_size, context_len, *features]."
        )
    if lookback_len > data.shape[1]:
        raise ShapeError(
            f"Invalid lookback_len={lookback_len} for data of shape {data.shape}."
        )
    if is_inference_data:
        if data.shape[1] != lookback_len:
            raise ShapeError(
                f"Invalid context length ({data.shape[1]}). For inference data, context window length should be matching the lookback length ({lookback_len})"
            )


# Sorting and parsing
class TopKReturnType(NamedTuple):
    values: np.ndarray
    indices: np.ndarray


def topk(vec: np.ndarray, k: int):
    assert np.ndim(vec) == 1, "'vec' must be a 1D array"
    assert k > 0, "k should be greater than 0"
    sort_perm = np.flip(np.argsort(vec))  # descending order
    indices = sort_perm[:k]
    values = vec[indices]
    return TopKReturnType(values, indices)


def fuzzy_parse_complex(vec: np.ndarray, tol: float = 10.0):
    assert issubclass(
        vec.dtype.type, np.complexfloating
    ), "The input element should be complex"
    rcond = tol * np.finfo(vec.dtype).eps
    pdist_real_part = pdist(vec.real[:, None])
    # Set the same element whenever pdist is smaller than eps*tol
    condensed_idxs = np.argwhere(pdist_real_part < rcond)[:, 0]
    fuzzy_real = vec.real.copy()
    if condensed_idxs.shape[0] >= 1:
        for idx in condensed_idxs:
            i, j = row_col_from_condensed_index(vec.real.shape[0], idx)
            avg = 0.5 * (fuzzy_real[i] + fuzzy_real[j])
            fuzzy_real[i] = avg
            fuzzy_real[j] = avg
    fuzzy_imag = vec.imag.copy()
    fuzzy_imag[np.abs(fuzzy_imag) < rcond] = 0.0
    return fuzzy_real + 1j * fuzzy_imag


def row_col_from_condensed_index(d, index):
    # Credits to: https://stackoverflow.com/a/14839010
    b = 1 - (2 * d)
    i = (-b - math.sqrt(b**2 - 8 * index)) // 2
    j = index + i * (b + i + 2) // 2 + 1
    return (int(i), int(j))


def parse_cplx_eig(vec: np.ndarray):
    _real_eigs_mask = np.abs(vec.imag) <= np.finfo(vec.dtype).eps
    real_eigs = vec[_real_eigs_mask].real
    _cplx_eigs_mask = np.logical_not(_real_eigs_mask)
    cplx_eigs = vec[_cplx_eigs_mask]
    cplx_conj_pairs_idxs = _parse_cplx_conj_pairs(cplx_eigs)
    return np.concatenate(
        [np.sort(real_eigs), np.sort(cplx_eigs[cplx_conj_pairs_idxs])]
    )


def _parse_cplx_conj_pairs(cplx_conj_vec: np.ndarray):
    if not cplx_conj_vec.shape[0] % 2 == 0:
        raise ValueError(
            f"The array must consist in a set of complex conjugate pairs, but its shape ({cplx_conj_vec.shape[0]}"
            f" is odd)."
        )
    _v_sort = np.argsort(cplx_conj_vec)
    _v_cj_sort = np.argsort(cplx_conj_vec.conj())

    _diff = cplx_conj_vec[_v_sort] - cplx_conj_vec.conj()[_v_cj_sort]
    if not np.allclose(_diff, np.zeros_like(_diff)):
        raise ValueError(
            "The provided array does not consists of complex conjugate pairs"
        )

    _v = cplx_conj_vec[_v_sort]

    idx_list = []
    for i in range(cplx_conj_vec.shape[0]):
        _idx_tuple = (_v_sort[i], _v_cj_sort[i])
        _idx_tuple_r = (_idx_tuple[1], _idx_tuple[0])
        if _idx_tuple in idx_list:
            continue
        elif _idx_tuple_r in idx_list:
            continue
        else:
            if np.angle(_v[i]) >= 0:
                idx_list.append(_idx_tuple)
            else:
                idx_list.append(_idx_tuple_r)
    _pos_phase_idxs = [i[0] for i in idx_list]
    return np.asarray(_pos_phase_idxs, dtype=int)



import numpy as np
import scipy.integrate

from kooplearn.datasets.misc import DataGenerator


class DuffingOscillator(DataGenerator):
    """
    A class for simulating the Duffing Oscillator.

    The Duffing Oscillator is a mathematical model used to describe a damped
    driven harmonic oscillator with nonlinear effects. It is commonly used
    in physics and engineering to study chaotic behavior.

    Args:
        alpha (float, optional): The stiffness coefficient (default is 0.5).
        beta (float, optional): The nonlinear coefficient (default is 0.0625).
        gamma (float, optional): The damping coefficient (default is 0.1).
        delta (float, optional): Another damping coefficient (default is 2.5).
        omega (float, optional): The angular frequency of the driving force (default is 2.0).
        dt (float, optional): The time step size for the numerical integration (default is 0.01).

    Attributes:
        alpha (float): The stiffness coefficient.
        beta (float): The nonlinear coefficient.
        gamma (float): The damping coefficient.
        delta (float): Another damping coefficient.
        omega (float): The angular frequency of the driving force.
        dt (float): The time step size for numerical integration.

    Examples:

    .. code-block:: python

        duffing = DuffingOscillator(alpha=0.5, beta=0.0625, gamma=0.1, delta=2.5, omega=2.0, dt=0.01)
        initial_conditions = np.array([0.0, 0.0])
        trajectory = duffing.sample(initial_conditions, T=100)

    """

    def __init__(
        self, alpha=0.5, beta=0.0625, gamma=0.1, delta=2.5, omega=2.0, dt=0.01
    ):
        """
        Initializes a DuffingOscillator object.

        Args:
            alpha (float, optional): The stiffness coefficient (default is 0.5).
            beta (float, optional): The nonlinear coefficient (default is 0.0625).
            gamma (float, optional): The damping coefficient (default is 0.1).
            delta (float, optional): Another damping coefficient (default is 2.5).
            omega (float, optional): The angular frequency of the driving force (default is 2.0).
            dt (float, optional): The time step size for the numerical integration (default is 0.01).
        """
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma
        self.delta = delta
        self.omega = omega
        self.dt = dt

    def D(self, t, x):
        """
        The derivative function representing the Duffing oscillator's equations of motion.

        Args:
            t (float): The current time.
            x (np.ndarray): An array representing the current state [position, velocity].

        Returns:
            np.ndarray: An array representing the derivatives of the state [velocity, acceleration].
        """
        dx = np.array(
            [
                x[1],
                -self.delta * x[1]
                - self.alpha * x[0]
                - self.beta * x[0] ** 3
                + self.gamma * np.cos(self.omega * t),
            ]
        )
        return dx

    def sample(self, X0: np.ndarray, T: int = 1):
        """
        Generate the trajectory of the Duffing oscillator.

        Args:
            X0 (np.ndarray): The initial conditions as an array [initial_position, initial_velocity].
            T (int, optional): The number of time steps (default is 1).

        Returns:
            np.ndarray: An array containing the trajectory of the oscillator with shape (T+1, 2),
                where each row represents [position, velocity] at a given time step.
        """
        sim_time = self.dt * (T + 1)
        t_eval = np.linspace(0, sim_time, T + 1, endpoint=True)
        t_span = (0, t_eval[-1])
        sol = scipy.integrate.solve_ivp(
            self.D, t_span, X0, t_eval=t_eval, method="RK45"
        )
        return sol.y.T


class Lorenz63(DataGenerator):
    """
    A class for simulating the Lorenz-63 chaotic dynamical system.

    The Lorenz-63 system is a simplified mathematical model of atmospheric
    convection that exhibits chaotic behavior.

    Args:
        sigma (float, optional): The :math:`\\sigma` parameter (default is 10).
        mu (float, optional): The :math:`\\mu` parameter (default is 28).
        beta (float, optional): The :math:`\\beta` parameter (default is 8/3).
        dt (float, optional): The time step size for numerical integration (default is 0.01).

    Attributes:
        sigma (float): The :math:`\\sigma` parameter.
        mu (float): The :math:`\\mu` parameter.
        beta (float): The :math:`\\beta` parameter.
        dt (float): The time step size for numerical integration.
        M_lin (np.ndarray): The linearized matrix of the Lorenz-63 system.

    Examples:

    .. code-block:: python

        lorenz = Lorenz63(sigma=10, mu=28, beta=8/3, dt=0.01)
        initial_conditions = np.array([1.0, 0.0, 0.0])
        trajectory = lorenz.sample(initial_conditions, T=100)

    """

    def __init__(self, sigma=10, mu=28, beta=8 / 3, dt=0.01):
        """
        Initializes a Lorenz63 object.

        Args:
            sigma (float, optional): The :math:`\\sigma` parameter (default is 10).
            mu (float, optional): The :math:`\\mu` parameter (default is 28).
            beta (float, optional): The :math:`\\beta` parameter (default is 8/3).
            dt (float, optional): The time step size for numerical integration (default is 0.01).
        """
        self.sigma = sigma
        self.mu = mu
        self.beta = beta
        self.dt = dt
        self.M_lin = np.array(
            [[-self.sigma, self.sigma, 0], [self.mu, 0, 0], [0, 0, -self.beta]]
        )

    def sample(self, X0: np.ndarray, T: int = 1):
        """
        Generate the trajectory of the Lorenz-63 system.

        Args:
            X0 (np.ndarray): The initial conditions as an array [x, y, z].
            T (int, optional): The number of time steps (default is 1).

        Returns:
            np.ndarray: An array containing the trajectory of the system with shape (T+1, 3),
                where each row represents [x, y, z] at a given time step.
        """
        sim_time = self.dt * (T + 1)
        t_eval = np.linspace(0, sim_time, T + 1, endpoint=True)
        t_span = (0, t_eval[-1])
        sol = scipy.integrate.solve_ivp(
            self.D, t_span, X0, t_eval=t_eval, method="RK45"
        )
        return sol.y.T

    def D(self, t, x):
        """
        The derivative function representing the Lorenz-63 equations.

        Args:
            t (float): The current time.
            x (np.ndarray): An array representing the current state [x, y, z].

        Returns:
            np.ndarray: An array representing the derivatives of the state [dx/dt, dy/dt, dz/dt].
        """
        dx = self.M_lin @ x
        dx[1] -= x[2] * x[0]
        dx[2] += x[0] * x[1]
        return dx



from pathlib import Path
from shutil import rmtree

import lightning
import numpy as np
import pytest
import torch

from kooplearn.data import traj_to_contexts
from kooplearn.datasets import Mock
from kooplearn.models import ConsistentAE, DynamicAE
from kooplearn.models.feature_maps import DPNet, VAMPNet
from kooplearn.nn.data import TrajToContextsDataset

NUM_SAMPLES = 100
DIM = 7

TRAJ = Mock(num_features=DIM, rng_seed=0).sample(None, NUM_SAMPLES).astype(np.float32)
DATA = traj_to_contexts(TRAJ)
NN_DATA = TrajToContextsDataset(TRAJ)
NN_DATALOADER = torch.utils.data.DataLoader(NN_DATA, batch_size=100, shuffle=True)


def _make_tmp_path(model_name: str):
    tmp_path = Path(__file__).parent / f"tmp/{model_name}.bin"
    return tmp_path


def _cleanup():
    tmp_path = Path(__file__).parent / "tmp"
    rmtree(tmp_path)


def _allclose(a, b):
    m = np.isfinite(a) & np.isfinite(b)
    return np.allclose(a[m], b[m], rtol=1e-3, atol=1e-5)


class Encoder(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.encoder = torch.nn.Linear(DIM, 32)

    def forward(self, x):
        return self.encoder(x)


class Decoder(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.decoder = torch.nn.Linear(32, DIM)

    def forward(self, x):
        return self.decoder(x)


def test_KernelDMD_serialization():
    from kooplearn.models import KernelDMD

    model = KernelDMD().fit(DATA)
    tmp_path = _make_tmp_path(model.__class__.__name__)
    model.save(tmp_path)
    restored_model = KernelDMD.load(tmp_path)

    # Check that predict, eig and modes return the same values
    assert _allclose(
        model.predict(DATA[:, : model.lookback_len, ...]),
        restored_model.predict(DATA[:, : model.lookback_len, ...]),
    )

    assert _allclose(
        model.modes(DATA[:, : model.lookback_len, ...]),
        restored_model.modes(DATA[:, : model.lookback_len, ...]),
    )

    e1, l1, r1 = model.eig(
        eval_left_on=DATA[:, : model.lookback_len, ...],
        eval_right_on=DATA[:, : model.lookback_len, ...],
    )
    e2, l2, r2 = restored_model.eig(
        eval_left_on=DATA[:, : model.lookback_len, ...],
        eval_right_on=DATA[:, : model.lookback_len, ...],
    )

    assert _allclose(e1, e2)
    assert _allclose(l1, l2)
    assert _allclose(r1, r2)

    _cleanup()


def test_ExtendedDMD_serialization():
    from kooplearn.models import ExtendedDMD

    model = ExtendedDMD().fit(DATA)
    tmp_path = _make_tmp_path(model.__class__.__name__)
    model.save(tmp_path)
    restored_model = ExtendedDMD.load(tmp_path)

    # Check that predict, eig and modes return the same values
    assert _allclose(
        model.predict(DATA[:, : model.lookback_len, ...]),
        restored_model.predict(DATA[:, : model.lookback_len, ...]),
    )

    assert _allclose(
        model.modes(DATA[:, : model.lookback_len, ...]),
        restored_model.modes(DATA[:, : model.lookback_len, ...]),
    )

    e1, l1, r1 = model.eig(
        eval_left_on=DATA[:, : model.lookback_len, ...],
        eval_right_on=DATA[:, : model.lookback_len, ...],
    )
    e2, l2, r2 = restored_model.eig(
        eval_left_on=DATA[:, : model.lookback_len, ...],
        eval_right_on=DATA[:, : model.lookback_len, ...],
    )

    assert _allclose(e1, e2)
    assert _allclose(l1, l2)
    assert _allclose(r1, r2)

    _cleanup()


@pytest.mark.parametrize("feature_map_cls", [DPNet, VAMPNet])
def test_DeepEDMD_serialization(feature_map_cls):
    from kooplearn.models import DeepEDMD

    trainer = lightning.Trainer(
        enable_progress_bar=False,
        enable_checkpointing=False,
        enable_model_summary=False,
        accelerator="cpu",
        max_epochs=1,
    )
    feature_map = feature_map_cls(
        Encoder, torch.optim.SGD, trainer, optimizer_kwargs={"lr": 1e-6}, seed=0
    )
    feature_map.fit(NN_DATALOADER)

    tmp_path = _make_tmp_path(feature_map.__class__.__name__)
    feature_map.save(tmp_path)
    restored_feature_map = feature_map_cls.load(tmp_path)

    assert _allclose(feature_map(TRAJ), restored_feature_map(TRAJ))

    model = DeepEDMD(feature_map=feature_map).fit(DATA)
    tmp_path = _make_tmp_path(model.__class__.__name__)
    model.save(tmp_path)
    restored_model = DeepEDMD.load(tmp_path)

    # Check that predict, eig and modes return the same values
    assert _allclose(
        model.predict(DATA[:, : model.lookback_len, ...]),
        restored_model.predict(DATA[:, : model.lookback_len, ...]),
    )

    assert _allclose(
        model.modes(DATA[:, : model.lookback_len, ...]),
        restored_model.modes(DATA[:, : model.lookback_len, ...]),
    )

    e1, l1, r1 = model.eig(
        eval_left_on=DATA[:, : model.lookback_len, ...],
        eval_right_on=DATA[:, : model.lookback_len, ...],
    )
    e2, l2, r2 = restored_model.eig(
        eval_left_on=DATA[:, : model.lookback_len, ...],
        eval_right_on=DATA[:, : model.lookback_len, ...],
    )

    assert _allclose(e1, e2)
    assert _allclose(l1, l2)
    assert _allclose(r1, r2)

    _cleanup()


@pytest.mark.parametrize("model_cls", [DynamicAE, ConsistentAE])
def test_AE_serialization(model_cls):
    trainer = lightning.Trainer(
        enable_progress_bar=False,
        enable_checkpointing=False,
        enable_model_summary=False,
        accelerator="cpu",
        max_epochs=1,
    )
    model = model_cls(
        Encoder,
        Decoder,
        32,
        torch.optim.SGD,
        trainer,
        optimizer_kwargs={"lr": 1e-6},
        seed=0,
    )
    model.fit(NN_DATALOADER)

    tmp_path = _make_tmp_path(model.__class__.__name__)
    model.save(tmp_path)
    restored_model = model_cls.load(tmp_path)

    # Check that predict, eig and modes return the same values
    assert _allclose(
        model.predict(DATA[:, : model.lookback_len, ...]),
        restored_model.predict(DATA[:, : model.lookback_len, ...]),
    )

    e1, l1, r1 = model.eig(
        eval_left_on=DATA[:, : model.lookback_len, ...],
        eval_right_on=DATA[:, : model.lookback_len, ...],
    )
    e2, l2, r2 = restored_model.eig(
        eval_left_on=DATA[:, : model.lookback_len, ...],
        eval_right_on=DATA[:, : model.lookback_len, ...],
    )

    assert _allclose(e1, e2)
    assert _allclose(l1, l2)
    assert _allclose(r1, r2)

    _cleanup()



def check_torch_deps():
    try:
        import lightning
    except ImportError:
        raise ImportError(
            "To use kooplearn's deep learning models please reinstall it with the `torch` extra flag by typing `pip install kooplearn[torch]`."
        )


def check_dashboard_deps():
    try:
        import dash_iconify
        import dash_mantine_components
        import plotly
    except ImportError:
        raise ImportError(
            "To use kooplearn's dashboard please reinstall it with the `dashboard` extra flag by typing `pip install kooplearn[dashboard]`."
        )



import functools
import os
import pathlib
import shutil
from warnings import warn

import requests
from tqdm.auto import tqdm

base_url = "http://ftp.imp.fu-berlin.de/pub/cmb-data/"
files = [
    "alanine-dipeptide-3x250ns-backbone-dihedrals.npz",
    "alanine-dipeptide-3x250ns-heavy-atom-distances.npz",
    "alanine-dipeptide-3x250ns-heavy-atom-positions.npz",
]


def download(url, filename):
    # Adapted from https://stackoverflow.com/questions/37573483/progress-bar-while-download-file-over-http-with-requests
    r = requests.get(url, stream=True, allow_redirects=True)
    if r.status_code != 200:
        r.raise_for_status()  # Will only raise for 4xx codes, so...
        raise RuntimeError(f"Request to {url} returned status code {r.status_code}")
    file_size = int(r.headers.get("Content-Length", 0))

    path = pathlib.Path(filename).expanduser().resolve()
    path.parent.mkdir(parents=True, exist_ok=True)

    desc = "(Unknown total file size)" if file_size == 0 else ""
    r.raw.read = functools.partial(
        r.raw.read, decode_content=True
    )  # Decompress if needed
    with tqdm.wrapattr(r.raw, "read", total=file_size, desc=desc) as r_raw:
        with path.open("wb") as f:
            shutil.copyfileobj(r_raw, f)
    return path


if __name__ == "__main__":
    example_path = pathlib.Path(__file__).parent.resolve()
    data_path = os.path.join(example_path, "__data__/")
    if not os.path.exists(data_path):
        os.makedirs(data_path)
    else:
        warn(
            "Warning: data folder already exists. Removing it and downloading new data."
        )
        shutil.rmtree(data_path)
        # Remove existing files
    for file in files:
        download(base_url + file, os.path.join(data_path, file))



import numpy as np


def directed_hausdorff_distance(pred: np.ndarray, reference: np.ndarray):
    """One-sided hausdorff distance between sets."""
    pred = np.asanyarray(pred)
    reference = np.asanyarray(reference)
    assert pred.ndim == 1
    assert reference.ndim == 1

    distances = np.zeros((pred.shape[0], reference.shape[0]), dtype=np.float64)
    for pred_idx, pred_pt in enumerate(pred):
        for reference_idx, reference_pt in enumerate(reference):
            distances[pred_idx, reference_idx] = np.abs(pred_pt - reference_pt)
    hausdorff_dist = np.max(np.min(distances, axis=1))
    return hausdorff_dist



import numpy as np
from typing import Optional

from kooplearn.datasets.misc import DataGenerator
from kooplearn.datasets.stochastic import DiscreteTimeDynamics

class DiscreteBlackScholes(DiscreteTimeDynamics):
    """
    Discretised simulation of the Black Scholes model.

    Args:
        A: transition matrix of the drift of shape ''(d, d)''.
        sigma1: covariance matrix of the increments of the prices brownian motion, shape is ''(d,d)''.
        dt: size of the increments.

    Attributes:
       sample: samples the process over T iterations with starting point X0. 
    """
    def __init__(self, A:np.ndarray, sigma:np.ndarray, dt:float, rng_seed: Optional[int]=None):
        self.dim = A.shape[0]
        self.A = A
        self.dt = dt
        self.sigma = sigma
        self.rng = np.random.default_rng(rng_seed)
    
    def _step(self, X:np.ndarray):
        return X + self.A @ X *self.dt + np.sqrt(self.dt) * X * self.rng.multivariate_normal(np.zeros(self.dim), self.sigma)

class DiscreteOhrnstein(DiscreteTimeDynamics):
    """
    Discretised simulation of the Ohrnstein-Ulhenbeck volatility model.

    Args:
        mu: mean of the volatility process, array of size ''d''.
        beta: driving coefficient of the volatility process, matrix of shape ''(d,d)''.
        sigma: covariance matrix of the increments of the volatility's volatility, matrix of shape ''(d,d)''.
        dt: size of the increments.

    Attributes:
       sample: samples the process over T iterations with starting point X0. 
    """
    def __init__(self, mu:np.ndarray, beta:np.ndarray, sigma:np.ndarray, dt:float, rng_seed:Optional[int]=None):
        self.dim = mu.shape[0]
        self.beta = beta
        self.mu = mu
        self.sigma = sigma
        self.dt = dt
        self.rng = np.random.default_rng(rng_seed)

    def _step(self, X:np.ndarray):
        return X + self.beta @ (self.mu - X) * self.dt + np.sqrt(self.dt) * self.rng.multivariate_normal(np.zeros(self.dim), self.sigma)

class DiscreteCIR(DiscreteOhrnstein):
    """
    Discretised simulation of the CIR volatility model.

    Args:
        mu: mean of the volatility process, array of size ''d''.
        beta: driving coefficient of the volatility process, matrix of shape ''(d,d)''.
        sigma: covariance matrix of the increments of the volatility's volatility, matrix of shape ''(d,d)''.
        dt: size of the increments.

    Attributes:
       sample: samples the process over T iterations with starting point X0. 
    """
    def _step(self, X:np.ndarray):
        return X + self.beta @ (self.mu - X)*self.dt + np.sqrt(self.dt)* np.sqrt(X) * self.rng.multivariate_normal(np.zeros(self.dim), self.sigma)

# Heston like model (ohrnstein volatility)
class DiscreteHeston(DiscreteTimeDynamics):
    """
    Discretised simulation of the Heston model.

    Args:
        A: transition matrix of the drift of shape ''(d, d)''.
        sigma1: covariance matrix of the increments of the prices brownian motion, shape is ''(d,d)''.
        mu: mean of the volatility process, array of size ''d''.
        beta: driving coefficient of the volatility process, matrix of shape ''(d,d)''.
        sigma2: covariance matrix of the increments of the volatility's volatility, matrix of shape ''(d,d)''.
        dt: size of the increments.

    Attributes:
       sample: samples the process over T iterations with starting point X0. 
    """
    def __init__(self, A:np.ndarray, 
                sigma1:np.ndarray,
                mu:np.ndarray, beta:np.ndarray, 
                sigma2:np.ndarray, 
                nu0:np.ndarray, 
                dt:float, 
                rng_seed:Optional[int]=None):

        self.A = A
        self.dim = A.shape[0]
        self.beta = beta
        self.mu = mu
        self.sigma1 = sigma1
        self.sigma2 = sigma2
        self.nu0 = nu0
        self.dt = dt
        self.rng = np.random.default_rng(rng_seed)
    def _step(self, X:np.ndarray):
        next_X = X + self.dt*self.A@X + np.sqrt(self.dt)*np.sqrt(self.nu0) * X * self.rng.multivariate_normal(np.zeros(self.dim), self.sigma2)
        self.nu0 += self.dt*self.beta @ (self.mu - self.nu0) + np.sqrt(self.dt)* np.sqrt(self.nu0) * self.rng.multivariate_normal(np.zeros(self.dim), self.sigma1)
        return next_X


class Garch(DataGenerator):
    # one dimensional one lag garch model
    def __init__(self, alpha:float, beta:float, alpha0:float=0.0):
        self.alpha = alpha
        self.beta = beta
        self.alpha0 = alpha0

    def sample(self, X0, T=1):
        memory = np.zeros(T + 1)
        memory[0] = X0
        aux_memory = np.zeros(T + 1)

        noise = np.random.normal(0, 1, size=T)

        for t in range(T):
            aux_memory[t + 1] = (
                self.alpha0 + self.alpha * (memory[t] ** 2) + self.beta * aux_memory[t]
            )
            memory[t + 1] = noise[t] * np.sqrt(aux_memory[t + 1])

        return memory


class DMgarch(DataGenerator):
    # mutlidimensional one lag diagonal mgarch model
    def __init__(self, s, A, B):
        self.A = A
        self.B = B
        self.s = s

    def sample(self, X0, T=1):
        memory = np.zeros((T, X0.shape))
        memory[0] = X0
        H = np.zeros((X0.shape, X0.shape))

        for t in range(T):
            H = self.s + self.A * np.outer(memory[t], memory[t]) + self.B * H
            memory[t + 1] = np.linalg.cholesky(H) @ np.random.normal(
                0, 1, size=X0.shape
            )

        return memory



from __future__ import annotations

import logging
from typing import Callable, Optional, Union

import numpy as np

from kooplearn._src.linalg import cov
from kooplearn._src.operator_regression import primal
from kooplearn._src.operator_regression.utils import (
    contexts_to_markov_train_states,
    parse_observables,
)
from kooplearn._src.serialization import pickle_load, pickle_save
from kooplearn._src.utils import ShapeError, check_contexts_shape, check_is_fitted
from kooplearn.abc import BaseModel, FeatureMap
from kooplearn.models.feature_maps import IdentityFeatureMap

logger = logging.getLogger("kooplearn")


class ExtendedDMD(BaseModel):
    """
    Extended Dynamic Mode Decomposition (ExtendedDMD) Model.
    Implements the ExtendedDMD estimators approximating the Koopman (deterministic systems) or Transfer (stochastic systems) operator following the approach described in :footcite:t:`Kostic2022`.

    Args:
        feature_map (callable): Dictionary of functions used for the ExtendedDMD algorithm. Should be a subclass of ``kooplearn.abc.FeatureMap``.
        reduced_rank (bool): If ``True`` initializes the reduced rank estimator introduced in :footcite:t:`Kostic2022`, if ``False`` initializes the classical principal component estimator.
        rank (int): Rank of the estimator. ``None`` returns the full rank estimator.
        tikhonov_reg (float): Tikhonov regularization coefficient. ``None`` is equivalent to ``tikhonov_reg = 0``, and internally calls specialized stable algorithms to deal with this specific case.
        svd_solver (str): Solver used to perform the internal SVD calcuations. Currently supported: `full`, uses LAPACK solvers, `arnoldi`, uses ARPACK solvers, `randomized`, uses randomized SVD algorithms as described in :footcite:t:`Turri2023`.
        iterated_power (int): Number of power iterations when using a randomized algorithm (``svd_solver == 'randomized'``).
        n_oversamples (int): Number of oversamples when using a randomized algorithm (``svd_solver == 'randomized'``).
        rng_seed (int): Random Number Generator seed. Only used when ``svd_solver == 'randomized'``.  Defaults to ``None``, that is no explicit seed is setted.

    .. tip::

        A powerful DMD variation proposed by :footcite:t:`Arbabi2017`, known as Hankel-DMD, evaluates the Koopman/Transfer estimators by stacking consecutive snapshots together in a Hankel matrix. When this model is fitted context windows of length > 2, the lookback window length is automatically set to ``context_len - 1``. Upon fitting, the whole lookback window is passed through the feature map and the results are then flattened and *concatenated* together, realizing an Hankel-EDMD estimator.

    Attributes:
        data_fit : Training data: array of context windows of shape ``(n_samples, context_len, *features_shape)``.
        cov_X : Covariance matrix of the feature map evaluated at the initial states, that is ``self.data_fit[:, :self.lookback_len, ...]``.
        cov_Y : Covariance matrix of the feature map evaluated at the evolved states, , that is ``self.data_fit[:, 1:self.lookback_len + 1, ...]``.
        cov_XY : Cross-covariance matrix between initial and evolved states.
        U : Projection matrix of shape (n_out_features, rank). The Koopman/Transfer operator is approximated as :math:`U U^T \mathrm{cov_{XY}}`.
    """

    def __init__(
        self,
        feature_map: FeatureMap = IdentityFeatureMap(),
        reduced_rank: bool = True,
        rank: Optional[int] = None,
        tikhonov_reg: Optional[float] = None,
        svd_solver: str = "full",
        iterated_power: int = 1,
        n_oversamples: int = 5,
        rng_seed: Optional[int] = None,
    ):
        # Perform checks on the input arguments:
        if svd_solver not in ["full", "arnoldi", "randomized"]:
            raise ValueError(
                "Invalid svd_solver. Allowed values are 'full', 'arnoldi' and 'randomized'."
            )
        if svd_solver == "randomized" and iterated_power < 0:
            raise ValueError("Invalid iterated_power. Must be non-negative.")
        if svd_solver == "randomized" and n_oversamples < 0:
            raise ValueError("Invalid n_oversamples. Must be non-negative.")
        self.rng_seed = rng_seed
        self._feature_map = feature_map
        self.rank = rank
        if tikhonov_reg is None:
            self.tikhonov_reg = 0.0
        else:
            self.tikhonov_reg = tikhonov_reg
        self.svd_solver = svd_solver
        self.iterated_power = iterated_power
        self.n_oversamples = n_oversamples
        self.reduced_rank = reduced_rank
        self._is_fitted = False
        self._lookback_len = -1

    @property
    def lookback_len(self) -> bool:
        return self._lookback_len

    @property
    def is_fitted(self) -> bool:
        return self._is_fitted

    def feature_map(self, X: np.ndarray):
        check_contexts_shape(X, self.lookback_len, is_inference_data=True)
        _trail_dims = X.shape[2:]
        _n_samples = X.shape[0]
        new_shape = (_n_samples * self.lookback_len, *_trail_dims)
        # Light wrapper around the true feature map to perform reshapings
        X = X.reshape(new_shape)
        feat_X = self._feature_map(X)
        _trail_dims = feat_X.shape[1:]
        feat_X = feat_X.reshape(_n_samples, self.lookback_len, *_trail_dims)
        return feat_X.reshape(_n_samples, -1)

    def fit(self, data: np.ndarray, verbose: bool = True) -> ExtendedDMD:
        """
        Fits the ExtendedDMD model using either a randomized or a non-randomized algorithm, and either a full rank or a reduced rank algorithm,
        depending on the parameters of the model.

        .. attention::

            If ``context_len = data.shape[1] > 2``, the attribute :attr:`lookback_len` will be automatically set to ``context_len - 1``. The feature map will be evaluated independently for each snapshot in the lookback window, and the results are concatenated together to form a single feature vector. The pseudo-code of this operation is

            .. code-block:: python

                X = data[:, :self.lookback_len, ...]
                n_samples = X.shape[0]
                trailing_dims = X.shape[2:]
                #Stack snapshots in the context window to evaluate the feature map
                X = X.reshape((n_samples*self.lookback_len, *trailing_dims))
                feat_X = self.feature_map(X).reshape(n_samples, self.lookback_len, -1)
                #Flatten and concatenate results
                feat_X = feat_X.reshape(n_samples, -1)

        Args:
            data (np.ndarray): Batch of context windows of shape ``(n_samples, context_len, *features_shape)``.

        Returns:
            The fitted estimator.
        """
        self._pre_fit_checks(data)
        if self.reduced_rank:
            if self.svd_solver == "randomized":
                vectors = primal.fit_rand_reduced_rank_regression(
                    self.cov_X,
                    self.cov_XY,
                    self.tikhonov_reg,
                    self.rank,
                    self.n_oversamples,
                    self.iterated_power,
                    self.rng_seed,
                )
            else:
                vectors = primal.fit_reduced_rank_regression(
                    self.cov_X,
                    self.cov_XY,
                    self.tikhonov_reg,
                    self.rank,
                    self.svd_solver,
                )
        else:
            if self.svd_solver == "randomized":
                vectors = primal.fit_rand_principal_component_regression(
                    self.cov_X,
                    self.tikhonov_reg,
                    self.rank,
                    self.n_oversamples,
                    self.iterated_power,
                )
            else:
                vectors = primal.fit_principal_component_regression(
                    self.cov_X, self.tikhonov_reg, self.rank, self.svd_solver
                )

        self.U = vectors
        assert self.U.shape[1] <= self.rank
        if self.U.shape[1] < self.rank:
            logger.warning(
                f"Warning: The fitting algorithm discarded {self.rank - self.U.shape[1]} dimensions of the {self.rank} requested out of numerical instabilities.\nThe rank attribute has been updated to {self.U.shape[1]}.\nConsider decreasing the rank parameter."
            )
            self.rank = self.U.shape[1]

        # Final Checks
        check_is_fitted(
            self, ["U", "cov_XY", "cov_X", "cov_Y", "data_fit", "lookback_len"]
        )
        self._is_fitted = True
        if verbose:
            print(
                f"Fitted {self.__class__.__name__} model. Lookback length set to {self.lookback_len}"
            )
        return self

    def risk(self, data: Optional[np.ndarray] = None) -> float:
        """Risk of the estimator on the validation ``data``.

        Args:
            data (np.ndarray or None): Batch of context windows of shape ``(n_samples, context_len, *features_shape)``. If ``None``, evaluates the risk on the training data.

        Returns:
            Risk of the estimator, see Equation 11 of :footcite:p:`Kostic2022` for more details.
        """
        if data is not None:
            check_contexts_shape(data, self.lookback_len)
            data = np.asanyarray(data)
            if data.shape[1] - 1 != self.lookback_len:
                raise ShapeError(
                    f"The  context length ({data.shape[1]}) of the validation data does not match the context length of the training data ({self.lookback_len + 1})."
                )
            X_fit, Y_fit = contexts_to_markov_train_states(data, self.lookback_len)
            cov_Xv, cov_Yv, cov_XYv = self._init_covs(X_fit, Y_fit)
        else:
            cov_Xv, cov_Yv, cov_XYv = self.cov_X, self.cov_Y, self.cov_XY

        return primal.estimator_risk(cov_Xv, cov_Yv, cov_XYv, self.cov_XY, self.U)

    def predict(
        self,
        data: np.ndarray,
        t: int = 1,
        observables: Optional[Callable] = None,
    ) -> np.ndarray:
        """
        Predicts the state or, if the system is stochastic, its expected value :math:`\mathbb{E}[X_t | X_0 = X]` after ``t`` instants given the initial conditions ``X = data[:, self.lookback_len:, ...]`` being the lookback slice of ``data``.

        .. attention::

            ``data.shape[1]`` must match the lookback length ``self.lookback_len``. Otherwise, an error is raised.

        If ``observables`` are not ``None``, returns the analogue quantity for the observable instead.

        Args:
            data (numpy.ndarray): Initial conditions to predict. Array of context windows with shape ``(n_init_conditions, self.lookback_len, *self.data_fit.shape[2:])`` (see the note above).
            t (int): Number of steps in the future to predict (returns the last one).
            observables (callable or None): Callable or ``None``. If callable should map batches of states of shape ``(batch, *self.data_fit.shape[2:])`` to batches of observables ``(batch, *obs_features_shape)``.

        Returns:
           The predicted (expected) state/observable at time :math:`t`, shape ``(n_init_conditions, *obs_features_shape)``.
        """
        check_is_fitted(
            self, ["U", "cov_XY", "cov_X", "cov_Y", "data_fit", "lookback_len"]
        )
        _obs, expected_shape, X_inference, X_fit = parse_observables(
            observables, data, self.data_fit, self.lookback_len
        )

        phi_Xin = self.feature_map(X_inference)
        phi_X = self.feature_map(X_fit)

        return (primal.predict(t, self.U, self.cov_XY, phi_Xin, phi_X, _obs)).reshape(
            expected_shape
        )

    def eig(
        self,
        eval_left_on: Optional[np.ndarray] = None,
        eval_right_on: Optional[np.ndarray] = None,
    ) -> Union[
        np.ndarray,
        tuple[np.ndarray, np.ndarray],
        tuple[np.ndarray, np.ndarray, np.ndarray],
    ]:
        """
        Returns the eigenvalues of the Koopman/Transfer operator and optionally evaluates left and right eigenfunctions.

        Args:
            eval_left_on (numpy.ndarray or None): Array of context windows on which the left eigenfunctions are evaluated, shape ``(n_samples, lookback_len, *features)``.
            eval_right_on (numpy.ndarray or None): Array of context windows on which the right eigenfunctions are evaluated, shape ``(n_samples, lookback_len, *features)``.

        Returns:
            Eigenvalues of the Koopman/Transfer operator, shape ``(rank,)``. If ``eval_left_on`` or ``eval_right_on``  are not ``None``, returns the left/right eigenfunctions evaluated at ``eval_left_on``/``eval_right_on``: shape ``(n_samples, rank)``.
        """

        check_is_fitted(self, ["U", "cov_XY", "lookback_len"])
        if hasattr(self, "_eig_cache"):
            w, vl, vr = self._eig_cache
        else:
            w, vl, vr = primal.estimator_eig(self.U, self.cov_XY)
            self._eig_cache = (w, vl, vr)

        if eval_left_on is None and eval_right_on is None:
            # (eigenvalues,)
            return w
        elif eval_left_on is None and eval_right_on is not None:
            # (eigenvalues, right eigenfunctions)
            check_contexts_shape(
                eval_right_on, self.lookback_len, is_inference_data=True
            )
            phi_Xin = self.feature_map(eval_right_on)
            return w, primal.evaluate_eigenfunction(phi_Xin, vr)
        elif eval_left_on is not None and eval_right_on is None:
            # (eigenvalues, left eigenfunctions)
            check_contexts_shape(
                eval_left_on, self.lookback_len, is_inference_data=True
            )
            phi_Xin = self.feature_map(eval_left_on)
            return w, primal.evaluate_eigenfunction(phi_Xin, vl)
        elif eval_left_on is not None and eval_right_on is not None:
            # (eigenvalues, left eigenfunctions, right eigenfunctions)
            check_contexts_shape(
                eval_right_on, self.lookback_len, is_inference_data=True
            )
            check_contexts_shape(
                eval_left_on, self.lookback_len, is_inference_data=True
            )
            phi_Xin_l = self.feature_map(eval_left_on)
            phi_Xin_r = self.feature_map(eval_right_on)

            return (
                w,
                primal.evaluate_eigenfunction(phi_Xin_l, vl),
                primal.evaluate_eigenfunction(phi_Xin_r, vr),
            )

    def modes(
        self,
        data: np.ndarray,
        observables: Optional[Callable] = None,
    ) -> np.ndarray:
        """
        Computes the mode decomposition of arbitrary observables of the Koopman/Transfer operator at the states defined by ``data``.

        Informally, if :math:`(\\lambda_i, \\xi_i, \\psi_i)_{i = 1}^{r}` are eigentriplets of the Koopman/Transfer operator, for any observable :math:`f` the i-th mode of :math:`f` at :math:`x` is defined as: :math:`\\lambda_i \\langle \\xi_i, f \\rangle \\psi_i(x)`. See :footcite:t:`Kostic2022` for more details.

        Args:
            data (numpy.ndarray): Initial conditions to compute the modes on. See :func:`predict` for additional details on the shape.
            observables (callable or None): Callable or ``None``. If callable should map batches of states of shape ``(batch, *self.data_fit.shape[2:])`` to batches of observables ``(batch, *obs_features_shape)``.
        Returns:
            Modes of the system at the states defined by ``data``. Array of shape ``(self.rank, n_samples, ...)``.
        """
        check_is_fitted(self, ["U", "data_fit", "cov_XY", "lookback_len", "data_fit"])
        _obs, expected_shape, X_inference, X_fit = parse_observables(
            observables, data, self.data_fit, self.lookback_len
        )

        phi_X = self.feature_map(X_fit)
        phi_Xin = self.feature_map(X_inference)

        _gamma = primal.estimator_modes(self.U, self.cov_XY, phi_X, phi_Xin)
        expected_shape = (self.rank,) + expected_shape
        return np.matmul(_gamma, _obs).reshape(
            expected_shape
        )  # [rank, num_initial_conditions, ...]

    def svals(self) -> np.ndarray:
        """Singular values of the Koopman/Transfer operator.

        Returns:
            The estimated singular values of the Koopman/Transfer operator. Array of shape `(rank,)`.
        """
        check_is_fitted(self, ["U", "cov_XY"])
        return primal.svdvals(self.U, self.cov_XY)

    def save(self, filename):
        """Serialize the model to a file.

        Args:
            filename (path-like or file-like): Save the model to file.
        """
        pickle_save(self, filename)

    @classmethod
    def load(cls, filename):
        """Load a serialized model from a file.

        Args:
            filename (path-like or file-like): Load the model from file.

        Returns:
            ExtendedDMD: The loaded model.
        """
        return pickle_load(cls, filename)

    def _init_covs(
        self, X: np.ndarray, Y: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Initializes the covariance matrices `cov_X`, `cov_Y`, and `cov_XY`.

        Args:
            stacked (np.ndarray): Training data of shape ``(n_samples, 2,  *features_shape)``. It should be the result of the function :func:`stack_lookback`.

        Returns:
            A tuple containing:
                - ``cov_X`` (np.ndarray): Covariance matrix of the feature map evaluated at X, shape ``(n_features, n_features)``.
                - ``cov_Y`` (np.ndarray): Covariance matrix of the feature map evaluated at Y, shape ``(n_features, n_features)``.
                - ``cov_XY`` (np.ndarray): Cross-covariance matrix of the feature map evaluated at X and Y, shape ``(n_features, n_features)``.
        """
        X = self.feature_map(X)
        Y = self.feature_map(Y)

        cov_X = cov(X)
        cov_Y = cov(Y)
        cov_XY = cov(X, Y)
        return cov_X, cov_Y, cov_XY

    def _pre_fit_checks(self, data: np.ndarray) -> None:
        """Performs pre-fit checks on the training data.

        Use :func:`check_contexts_shape` to check and sanitize the input data, initialize the covariance matrices and saves the training data.

        Args:
            data (np.ndarray): Batch of context windows of shape ``(n_samples, context_len, *features_shape)``.
        """
        lookback_len = data.shape[1] - 1
        check_contexts_shape(data, lookback_len)
        data = np.asanyarray(data)

        # Save the lookback length
        self._lookback_len = lookback_len
        X_fit, Y_fit = contexts_to_markov_train_states(data, self.lookback_len)

        self.cov_X, self.cov_Y, self.cov_XY = self._init_covs(X_fit, Y_fit)
        self.data_fit = data

        if self.rank is None:
            self.rank = self.cov_X.shape[0]
            logger.info(f"Rank of the estimator set to {self.rank}")

        if hasattr(self, "_eig_cache"):
            del self._eig_cache


class DMD(ExtendedDMD):
    """
    Dynamic Mode Decomposition (DMD) Model.
    Implements the classical DMD estimator approximating the Koopman (deterministic systems) or Transfer (stochastic systems) operator. This model just a minimal wrapper around ``ExtendedDMD`` setting the feature map to the identity function.
    """

    def __init__(
        self,
        reduced_rank: bool = True,
        rank: Optional[int] = 5,
        tikhonov_reg: float = 0,
        svd_solver: str = "full",
        iterated_power: int = 1,
        n_oversamples: int = 5,
        rng_seed: Optional[int] = None,
    ):
        super().__init__(
            reduced_rank=reduced_rank,
            rank=rank,
            tikhonov_reg=tikhonov_reg,
            svd_solver=svd_solver,
            iterated_power=iterated_power,
            n_oversamples=n_oversamples,
            rng_seed=rng_seed,
        )



import numpy as np
import pytest

from kooplearn.data import traj_to_contexts
from kooplearn.nn.data import TrajToContextsDataset


@pytest.mark.parametrize("context_window_len", [-1, 0, 1, 2, 3, 4])
@pytest.mark.parametrize("time_lag", [-1, 0, 1, 2, 3, 4])
@pytest.mark.parametrize("trj_len", [1, 2, 10, 50])
@pytest.mark.parametrize("n_feature_axes", [0, 1, 2, 3])
def test_traj_to_contexts(context_window_len, time_lag, trj_len, n_feature_axes):
    trj = np.arange(trj_len)
    if n_feature_axes > 0:
        trj = np.reshape(trj, (trj_len,) + (1,) * n_feature_axes)
        assert trj.shape == (trj_len,) + (1,) * n_feature_axes
    else:
        assert trj.shape == (trj_len,)
        n_feature_axes = 1
    _C = 1 + (context_window_len - 1) * time_lag

    if context_window_len < 2:
        with pytest.raises(ValueError):
            _ = traj_to_contexts(trj, context_window_len, time_lag)
        return

    if time_lag < 1:
        with pytest.raises(ValueError):
            _ = traj_to_contexts(trj, context_window_len, time_lag)
        return

    if _C > trj_len:
        with pytest.raises(ValueError):
            _ = traj_to_contexts(trj, context_window_len, time_lag)
        return

    res = traj_to_contexts(trj, context_window_len, time_lag)
    assert res.shape == (trj_len - _C + 1, context_window_len) + (1,) * n_feature_axes
    if n_feature_axes == 1:
        for i in range(res.shape[1]):
            assert np.all(res[:, i, 0] == np.arange(trj_len - _C + 1) + i * time_lag)


@pytest.mark.parametrize("context_window_len", [-1, 0, 1, 2, 3, 4])
@pytest.mark.parametrize("time_lag", [-1, 0, 1, 2, 3, 4])
@pytest.mark.parametrize("trj_len", [1, 2, 10, 50])
@pytest.mark.parametrize("n_feature_axes", [0, 1, 2, 3])
def test_torch_traj_to_contexts(context_window_len, time_lag, trj_len, n_feature_axes):
    trj = np.arange(trj_len)
    if n_feature_axes > 0:
        trj = np.reshape(trj, (trj_len,) + (1,) * n_feature_axes)
        assert trj.shape == (trj_len,) + (1,) * n_feature_axes
    else:
        assert trj.shape == (trj_len,)
        n_feature_axes = 1
    _C = 1 + (context_window_len - 1) * time_lag

    if context_window_len < 2:
        with pytest.raises(ValueError):
            _ = TrajToContextsDataset(trj, context_window_len, time_lag)
        return

    if time_lag < 1:
        with pytest.raises(ValueError):
            _ = TrajToContextsDataset(trj, context_window_len, time_lag)
        return

    if _C > trj_len:
        with pytest.raises(ValueError):
            _ = TrajToContextsDataset(trj, context_window_len, time_lag)
        return

    res = (
        TrajToContextsDataset(trj, context_window_len, time_lag)
        .contexts.detach()
        .cpu()
        .numpy()
    )
    assert res.shape == (trj_len - _C + 1, context_window_len) + (1,) * n_feature_axes
    if n_feature_axes == 1:
        for i in range(res.shape[1]):
            assert np.all(res[:, i, 0] == np.arange(trj_len - _C + 1) + i * time_lag)



from __future__ import annotations

import logging
from typing import Optional

from kooplearn._src.serialization import pickle_load, pickle_save
from kooplearn._src.utils import NotFittedError
from kooplearn.abc import TrainableFeatureMap
from kooplearn.models import ExtendedDMD

logger = logging.getLogger("kooplearn")


class DeepEDMD(ExtendedDMD):
    """
    Deep Extended Dynamic Mode Decomposition (DeepEDMD) Model.

    Implements the Extended Dynamic Mode Decomposition estimators approximating the Koopman (deterministic systems) or Transfer (stochastic systems) operator following the approach described in :footcite:t:`Kostic2022`. In Deep Extended Dynamic Mode Decomposition the feature map used to embed the data is learned as well. :guilabel:`TODO - Add different refs`

    This model implements every method of :class:`kooplearn.models.ExtendedDMD`.

    .. caution::

        The feature map passed as a first argument should be already trained, that is ``feature_map.is_fitted == True``. If this is not the case, a ``NotFittedError`` is raised.

    Args:
        feature_map (callable): *Trained* feature map used for the DeepEDMD algorithm. Should be a subclass of :class:`kooplearn.abc.TrainableFeatureMap`.
        reduced_rank (bool): If ``True`` initializes the reduced rank estimator introduced in :footcite:t:`Kostic2022`, if ``False`` initializes the classical principal component estimator.
        rank (int): Rank of the estimator. ``None`` returns the full rank estimator.
        tikhonov_reg (float): Tikhonov regularization coefficient. ``None`` is equivalent to ``tikhonov_reg = 0``, and internally calls specialized stable algorithms to deal with this specific case.
        svd_solver (str): Solver used to perform the internal SVD calcuations. Currently supported: `full`, uses LAPACK solvers, `arnoldi`, uses ARPACK solvers, `randomized`, uses randomized SVD algorithms as described in :guilabel:`TODO - ADD REF`.
        iterated_power (int): Number of power iterations when using a randomized algorithm (``svd_solver == 'randomized'``).
        n_oversamples (int): Number of oversamples when using a randomized algorithm (``svd_solver == 'randomized'``).
        rng_seed (int): Random Number Generator seed. Only used when ``svd_solver == 'randomized'``.  Defaults to ``None``, that is no explicit seed is setted.
    """

    def __init__(
        self,
        feature_map: TrainableFeatureMap,
        reduced_rank: bool = True,
        rank: Optional[int] = None,
        tikhonov_reg: Optional[float] = None,
        svd_solver: str = "full",
        iterated_power: int = 1,
        n_oversamples: int = 5,
        rng_seed: Optional[int] = None,
    ):
        # Check that the provided feature map is trainable
        assert hasattr(
            feature_map, "fit"
        ), "The provided feature map is not trainable. Please provide a subclass of kooplearn.abc.TrainableFeatureMap."
        if not feature_map.is_fitted:
            raise NotFittedError(
                """
                The provided feature map is not fitted. Please call the fit method before initializing the DeepEDMD model.
                """
            )
        super().__init__(
            feature_map=feature_map,
            reduced_rank=reduced_rank,
            rank=rank,
            tikhonov_reg=tikhonov_reg,
            svd_solver=svd_solver,
            iterated_power=iterated_power,
            n_oversamples=n_oversamples,
            rng_seed=rng_seed,
        )

    @property
    def is_fitted(self) -> bool:
        return self._is_fitted

    def save(self, filename):
        """Serialize the model to a file.

        Args:
            filename (path-like or file-like): Save the model to file.
        """
        pickle_save(self, filename)

    @classmethod
    def load(cls, filename):
        """Load a serialized model from a file.

        Args:
            filename (path-like or file-like): Load the model from file.

        Returns:
            DeepEDMD: The loaded model.
        """
        return pickle_load(cls, filename)



import logging

from kooplearn._src.check_deps import check_torch_deps
from kooplearn._src.utils import ShapeError

logger = logging.getLogger("kooplearn")
check_torch_deps()
import torch  # noqa: E402
from torch.utils.data import Dataset  # noqa: E402


class ContextsDataset(Dataset):
    """A torch ``Dataset`` of context windows. This class is also aliased in :mod:`kooplearn.data` as :class:`kooplearn.data.ContextsDataset`.

    Args:
        contexts (torch.Tensor): A tensor of context windows.
    """

    def __init__(self, contexts: torch.Tensor):
        if not isinstance(contexts, torch.Tensor):
            logger.warning(
                f"The provided contexts are of type {type(contexts)}. Converting to torch.Tensor."
            )
            contexts = torch.as_tensor(contexts)
        if contexts.ndim < 2:
            raise ShapeError(
                f"Invalid contexts shape {contexts.shape}. The contexts must be at least three dimensional [num_samples, context_len, *features]."
            )
        self.contexts = contexts

    def __len__(self):
        return len(self.contexts)

    def __getitem__(self, idx):
        return self.contexts[idx]


def traj_to_contexts_dataset(
    trajectory: torch.Tensor, context_window_len: int = 2, time_lag: int = 1
):
    """Convert a single trajectory to a Torch Dataset of context windows. This function is also aliased in :mod:`kooplearn.data` as :func:`kooplearn.data.traj_to_contexts_dataset`.

    Args:
        trajectory (torch.tensor): A trajectory of shape ``(n_frames, *features_shape)``.
        context_window_len (int, optional): Length of the context window. Defaults to 2.
        time_lag (int, optional): Time lag, i.e. stride, between successive context windows. Defaults to 1.

    Returns:
        An instance of :class:`kooplearn.nn.data.ContextsDataset`.
    """

    if context_window_len < 2:
        raise ValueError(f"context_window_len must be >= 2, got {context_window_len}")
    if time_lag < 1:
        raise ValueError(f"time_lag must be >= 1, got {time_lag}")
    if trajectory.ndim == 0:
        trajectory = trajectory.reshape(1, 1)
    elif trajectory.ndim == 1:
        trajectory = trajectory[:, None]

    if not isinstance(trajectory, torch.Tensor):
        logger.warning(
            f"The provided trajectory is of type {type(trajectory)}. Converting to torch.Tensor."
        )
        trajectory = torch.as_tensor(trajectory)

    _context_window_len = 1 + (context_window_len - 1) * time_lag
    if _context_window_len > trajectory.shape[0]:
        raise ValueError(
            f"Invalid combination of context_window_len={context_window_len} and time_lag={time_lag} for trajectory of length {trajectory.shape[0]}. Try reducing context_window_len or time_lag."
        )

    contexts = trajectory.unfold(0, _context_window_len, 1)
    contexts = torch.movedim(contexts, -1, 1)[:, ::time_lag, ...]
    return ContextsDataset(contexts)


class TrajToContextsDataset(ContextsDataset):
    def __init__(
        self, trajectory: torch.Tensor, context_window_len: int = 2, time_lag: int = 1
    ):
        if context_window_len < 2:
            raise ValueError(
                f"context_window_len must be >= 2, got {context_window_len}"
            )
        if time_lag < 1:
            raise ValueError(f"time_lag must be >= 1, got {time_lag}")
        if trajectory.ndim == 0:
            trajectory = trajectory.reshape(1, 1)
        elif trajectory.ndim == 1:
            trajectory = trajectory[:, None]

        if not isinstance(trajectory, torch.Tensor):
            logger.warning(
                f"The provided trajectory is of type {type(trajectory)}. Converting to torch.Tensor."
            )
            trajectory = torch.as_tensor(trajectory)

        _context_window_len = 1 + (context_window_len - 1) * time_lag
        if _context_window_len > trajectory.shape[0]:
            raise ValueError(
                f"Invalid combination of context_window_len={context_window_len} and time_lag={time_lag} for trajectory of length {trajectory.shape[0]}. Try reducing context_window_len or time_lag."
            )

        self.context_len = context_window_len
        self.time_lag = time_lag
        self.contexts = trajectory.unfold(0, _context_window_len, 1)
        self.contexts = torch.movedim(self.contexts, -1, 1)[:, :: self.time_lag, ...]



import logging
import math
from pathlib import Path
from typing import Optional

import numpy as np
import scipy
import scipy.sparse
from scipy.integrate import romb
from scipy.special import binom
from scipy.stats.sampling import NumericalInversePolynomial

from kooplearn._src.utils import topk
from kooplearn.datasets.misc import (
    DataGenerator,
    DiscreteTimeDynamics,
    LinalgDecomposition,
)

logger = logging.getLogger("kooplearn")

try:
    import sdeint

    _has_sdeint = True
except ImportError:
    _has_sdeint = False


class Mock(DataGenerator):
    def __init__(self, num_features: int = 50, rng_seed: Optional[int] = None):
        self.rng = np.random.default_rng(rng_seed)
        self.num_features = num_features

    def sample(self, X0: np.ndarray, T: int = 1):
        return self.rng.random((T + 1, self.num_features))


class LinearModel(DiscreteTimeDynamics):
    def __init__(
        self, A: np.ndarray, noise: float = 0.0, rng_seed: Optional[int] = None
    ):
        self.A = A
        self.noise = noise
        self.rng = np.random.default_rng(rng_seed)

    def _step(self, X: np.ndarray):
        return self.A @ X + self.noise * self.rng.standard_normal(size=X.shape)


class RegimeChangeVAR(DiscreteTimeDynamics):
    def __init__(
        self,
        phi1: np.ndarray,
        phi2: np.ndarray,
        transition: np.ndarray,
        noise: float = 0.0,
        rng_seed: Optional[int] = None,
    ):
        self.phi1 = phi1
        self.phi2 = phi2
        self.transition = transition
        self.noise = noise
        self.rng = np.random.default_rng(rng_seed)
        self.current_state = 0

    def _step(self, X: np.ndarray):
        rand_trans = np.random.uniform(0, 1)
        if rand_trans < self.transition[self.current_state, 0]:
            self.current_state = 0
            return self.phi1 @ X + self.noise * self.rng.standard_normal(size=X.shape)
        else:
            self.current_state = 1
            return self.phi2 @ X + self.noise * self.rng.standard_normal(size=X.shape)


# Noisy Logistic Map
class CosineDistribution:
    def __init__(self, N):
        self.N = N
        self.C_N = np.pi / scipy.special.beta(N // 2 + 0.5, 0.5)

    def pdf(self, x):
        return self.C_N * ((np.cos(np.pi * x)) ** self.N)


class LogisticMap(DiscreteTimeDynamics):
    """Noisy Logistic map

    Args:
        r (float): parameter of the logistic map. Defaults to 4.0.
        N (int): Exponent of the trigonometric noise as defined in :footcite:t:`Kostic2022` (Appendix). Should be an _even_ integer or None (no noise). Defaults to None.
        rng_seed (int): Internal number generator seed. Defaults to None.

    """

    def __init__(
        self, r: float = 4.0, N: Optional[int] = None, rng_seed: Optional[int] = None
    ):
        self.rng_seed = rng_seed
        self.r = r
        self._rng = np.random.default_rng(self.rng_seed)
        if N is not None:
            # Noisy logistic map
            if N % 2 != 0:
                raise ValueError("N must be even")
            self.has_noise = True
            self.N = N
            self._noise_rng = NumericalInversePolynomial(
                CosineDistribution(N),
                domain=(-0.5, 0.5),
                mode=0,
                random_state=self._rng,
            )
            (
                self._svdvals,
                self._eigvals,
                self._lv,
                self._rv,
            ) = self._init_transfer_matrices()

        else:
            # Noiseless logistic map
            self.has_noise = False

    def predict(self, X_init: np.ndarray, T: int = 1):
        raise NotImplementedError("This method is not implemented yet")

    def eig(
        self,
        eval_left_on: Optional[np.ndarray] = None,
        eval_right_on: Optional[np.ndarray] = None,
    ):
        if not self.has_noise:
            raise ValueError("This method is only available for the noisy logistic map")
        perron_eig_idx = np.argmax(np.abs(self._eigvals))

        if eval_left_on is None and eval_right_on is None:  # None
            return self._eigvals
        elif eval_left_on is not None and eval_right_on is None:  # Only left
            Xl = np.asanyarray(eval_left_on).reshape(-1)
            betas_mat = np.stack(
                [self.noise_feature(Xl, i) for i in range(self.N + 1)], axis=1
            )  # [Xl.shape[0], N + 1]

            lfuncs = betas_mat @ self._lv
            # Standardize sign
            perron_fun_sign = np.sign(np.sign(lfuncs[:, perron_eig_idx]).mean())
            return self._eigvals, lfuncs * perron_fun_sign
        elif eval_left_on is None and eval_right_on is not None:  # Only right
            Xr = np.asanyarray(eval_right_on).reshape(-1)
            alphas_mat = np.stack(
                [self.noise_feature_composed_map(Xr, i) for i in range(self.N + 1)],
                axis=1,
            )  # [Xr.shape[0], N + 1]
            rfuncs = alphas_mat @ self._rv
            perron_fun_sign = np.sign(np.sign(rfuncs[:, perron_eig_idx]).mean())
            return self._eigvals, rfuncs * perron_fun_sign

        else:  # All
            # Left
            Xl = np.asanyarray(eval_left_on).reshape(-1)
            betas_mat = np.stack(
                [self.noise_feature(Xl, i) for i in range(self.N + 1)], axis=1
            )  # [Xl.shape[0], N + 1]
            lfuncs = betas_mat @ self._lv
            # Right
            Xr = np.asanyarray(eval_right_on).reshape(-1)
            alphas_mat = np.stack(
                [self.noise_feature_composed_map(Xr, i) for i in range(self.N + 1)],
                axis=1,
            )  # [Xr.shape[0], N + 1]
            rfuncs = alphas_mat @ self._rv
            perron_fun_sign = np.sign(np.sign(rfuncs[:, perron_eig_idx]).mean())
            return self._eigvals, lfuncs * perron_fun_sign, rfuncs * perron_fun_sign

    def svals(self):
        if not self.has_noise:
            raise ValueError("This method is only available for the noisy logistic map")
        return self._svdvals

    def _step(self, X_0: np.ndarray):
        return self.map(X_0)

    def map(self, X_init: np.ndarray, noisy: Optional[bool] = None):
        if noisy is None:
            noisy = self.has_noise
        # Dim checks on X_init
        if X_init.ndim == 0:
            X_init = X_init.reshape(1, 1)
        elif X_init.ndim == 1:
            X_init = X_init.reshape(-1, 1)
        elif X_init.ndim > 2:
            raise ValueError("X_init must be a scalar, a 1D array or a 2D array")
        if noisy:
            y = self.r * X_init * (1 - X_init)
            assert hasattr(
                self, "_noise_rng"
            ), "Noise rng not initialized, initialize the logistic map with N != None"
            xi = self._noise_rng.rvs(X_init.shape)
            return np.mod(y + xi, 1)
        else:
            return self.r * X_init * (1 - X_init)

    def noise_feature(self, x, i):  # beta
        if not self.has_noise:
            raise ValueError("This method is only available for the noisy logistic map")
        N = self.N
        normalization_cst = np.pi / scipy.special.beta(N // 2 + 0.5, 0.5)
        return (
            ((np.sin(np.pi * x)) ** (N - i))
            * ((np.cos(np.pi * x)) ** i)
            * np.sqrt(binom(N, i) * normalization_cst)
        )

    def noise_feature_composed_map(self, x, i):  # alpha = beta circ logistic
        r = self.r
        return self.noise_feature(r * x * (1 - x), i)

    def _init_transfer_matrices(self):
        if not self.has_noise:
            raise ValueError("This method is only available for the noisy logistic map")
        N = self.N

        def alphas_mat_el(i: int, j: int):
            def pairing(x):
                return self.noise_feature_composed_map(
                    x, i
                ) * self.noise_feature_composed_map(x, j)

            return scipy.integrate.quad(pairing, 0, 1)[0]

        def betas_mat_el(i: int, j: int):
            def pairing(x):
                return self.noise_feature(x, i) * self.noise_feature(x, j)

            return scipy.integrate.quad(pairing, 0, 1)[0]

        def koopman_el(i: int, j: int):
            def pairing(x):
                return self.noise_feature(x, i) * self.noise_feature_composed_map(x, j)

            return scipy.integrate.quad(pairing, 0, 1)[0]

        K = np.array([[koopman_el(i, j) for j in range(N + 1)] for i in range(N + 1)])
        A = np.array(
            [[alphas_mat_el(i, j) for j in range(N + 1)] for i in range(N + 1)]
        )
        B = np.array([[betas_mat_el(i, j) for j in range(N + 1)] for i in range(N + 1)])

        eigvals, lv, rv = scipy.linalg.eig(K, left=True, right=True)
        svdvals = scipy.linalg.eigvals(B @ A)
        # Check that svdvals are real and positive
        if not np.all(np.isreal(svdvals)):
            logger.warning("Singular values are not real, taking the real part")
        svdvals = svdvals.real

        if min(svdvals) < -np.finfo(svdvals.dtype).eps * len(svdvals):
            logger.warning(
                "Singular values are not positive, truncating to positive values"
            )
            svdvals = np.where(svdvals < 0, 0, svdvals)

        return np.sqrt(svdvals), eigvals, lv, rv


class MullerBrownPotential(DataGenerator):
    def __init__(self, kt: float = 1.5e3, rng_seed: Optional[int] = None):
        if not _has_sdeint:
            raise ImportError("sdeint not found, please install it to use this class")
        self.a = np.array([-1, -1, -6.5, 0.7])
        self.b = np.array([0.0, 0.0, 11.0, 0.6])
        self.c = np.array([-10, -10, -6.5, 0.7])
        self.A = np.array([-200, -100, -170, 15])
        self.X = np.array([1, 0, -0.5, -1])
        self.Y = np.array([0, 0.5, 1.5, 1])
        self.kt = kt
        self.rng = np.random.default_rng(rng_seed)

    def sample(self, X0: np.ndarray, T: int = 1):
        tspan = np.arange(0, 0.1 * T, 0.1)
        result = sdeint.itoint(
            self.neg_grad_potential, self.noise_term, X0, tspan, self.rng
        )
        return result

    def potential(self, X: np.ndarray):
        t1 = X[0] - self.X
        t2 = X[1] - self.Y
        tinexp = self.a * (t1**2) + self.b * t1 * t2 + self.c * (t2**2)
        return np.dot(self.A, np.exp(tinexp))

    def neg_grad_potential(self, x: np.ndarray, t: np.ndarray):
        """
        A_j * exp[a_j * (x1^2 - 2 * x1 * X_j) + b_j * (x1 * x2 - x1 * Y_j)]

        -> grad inner exp:
            a_j * (2 * x1 - 2 * X_j) + b_j * (x2 - Y_j)
        """
        t1 = x[0] - self.X
        t2 = x[1] - self.Y
        grad_inner_exp_x1 = 2 * self.a * t1 + self.b * t2
        grad_inner_exp_x2 = 2 * self.c * t2 + self.b * t1
        tinexp = np.exp(self.a * (t1**2) + self.b * t1 * t2 + self.c * (t2**2))

        return (
            -np.array(
                [
                    np.dot(self.A, tinexp * grad_inner_exp_x1),
                    np.dot(self.A, tinexp * grad_inner_exp_x2),
                ]
            )
            / self.kt
        )

    def noise_term(self, x: np.ndarray, t: np.ndarray):
        return np.diag([math.sqrt(2 * 1e-2), math.sqrt(2 * 1e-2)])


class LangevinTripleWell1D(DiscreteTimeDynamics):
    def __init__(
        self,
        gamma: float = 0.1,
        kt: float = 1.0,
        dt: float = 1e-4,
        rng_seed: Optional[int] = None,
    ):
        self.gamma = gamma
        self.kt = kt
        self.rng = np.random.default_rng(rng_seed)
        self.dt = dt

        self._inv_gamma = (self.gamma) ** -1

    def eig(self):
        if not hasattr(self, "_ref_evd"):
            self._compute_ref_evd()
        return self._ref_evd

    def _compute_ref_evd(self):
        assets_path = Path(__file__).parent / "assets"

        lap_x = scipy.sparse.load_npz(assets_path / "1D_triple_well_lap_x.npz")
        grad_x = scipy.sparse.load_npz(assets_path / "1D_triple_well_grad_x.npz")
        cc = np.load(assets_path / "1D_triple_well_cc.npz")

        force = scipy.sparse.diags(self.force_fn(cc))

        # Eq.(31) of https://doi.org/10.1007/978-3-642-56589-2_9 recalling that \sigma^2/(\gamma*2) = kBT
        generator = self._inv_gamma * self.kt * lap_x + self._inv_gamma * force.dot(
            grad_x
        )
        generator = generator * self.dt

        vals, vecs = np.linalg.eig(generator.toarray())

        # Filter out timescales smaller than dt
        mask = np.abs(vals.real) < 1.0
        vals = vals[mask]
        vecs = vecs[:, mask]

        vals = np.exp(vals)
        # Checking that the eigenvalues are real
        type_ = vals.dtype.type
        f = np.finfo(type_).eps

        tol = f * 1000
        if not np.all(np.abs(vals.imag) < tol):
            raise ValueError(
                "The computed eigenvalues are not real, try to decrease dt"
            )
        else:
            vals = vals.real

        _k = len(vals)
        evd_sorting_perm = topk(vals, _k)
        vals = evd_sorting_perm.values
        vecs = vecs[:, evd_sorting_perm.indices].real

        dx = cc[1] - cc[0]
        boltzmann_pdf = (dx**-1) * scipy.special.softmax(self.force_fn(cc) / self.kt)
        abs2_eigfun = (np.abs(vecs) ** 2).T
        eigfuns_norms = np.sqrt(romb(boltzmann_pdf * abs2_eigfun, dx=dx, axis=-1))
        vecs = vecs * (eigfuns_norms**-1.0)

        # Storing the results
        self._ref_evd = LinalgDecomposition(vals, cc, vecs)
        self._ref_boltzmann_density = boltzmann_pdf

    def _step(self, X: np.ndarray):
        F = self.force_fn(X)
        xi = self.rng.standard_normal(X.shape)
        dX = (
            F * self._inv_gamma * self.dt
            + np.sqrt(2.0 * self.kt * self.dt * self._inv_gamma) * xi
        )
        return X + dX

    def force_fn(self, x: np.ndarray):
        return -1.0 * (
            -128 * np.exp(-80 * ((-0.5 + x) ** 2)) * (-0.5 + x)
            - 512 * np.exp(-80 * (x**2)) * x
            + 32 * (x**7)
            - 160 * np.exp(-40 * ((0.5 + x) ** 2)) * (0.5 + x)
        )

    def _eigfun_sign_phase(self, estimated, true):
        norm_p = np.linalg.norm(estimated + true)
        norm_m = np.linalg.norm(estimated - true)
        if norm_p <= norm_m:
            return -1.0
        else:
            return 1.0

    def _standardize_evd(
        self, evd: LinalgDecomposition, dx: float, density: Optional[np.ndarray] = None
    ) -> LinalgDecomposition:
        # Sorting and normalizing
        sort_perm = np.flip(np.argsort(evd.values.real))
        functions = (evd.vectors[:, sort_perm]).real
        abs2_eigfun = (np.abs(functions) ** 2).T
        if density is not None:
            abs2_eigfun *= density
        # Norms
        funcs_norm = np.sqrt(romb(abs2_eigfun, dx=dx, axis=-1))
        functions *= funcs_norm**-1.0
        values = (evd.values.real)[sort_perm]
        return LinalgDecomposition(values, functions)

    def standardize_eigenfunction_phase(
        self, evd: LinalgDecomposition
    ) -> LinalgDecomposition:
        assert np.allclose(evd.x, self._ref_domain_sample)

        ref_evd = self._ref_evd
        density = self._ref_boltzmann_density
        dx = self._ref_domain_sample[1] - self._ref_domain_sample[0]

        # ref_evd is assumed already standardized
        evd = self._standardize_evd(evd, dx, density=density)
        phase_aligned_funcs = evd.vectors.copy()
        num_funcs = evd.vectors.shape[1]
        for r in range(num_funcs):
            estimated = evd.vectors[:, r]
            true = ref_evd.vectors[:, r]
            phase_aligned_funcs[:, r] = (
                self._eigfun_sign_phase(estimated * density, true * density) * estimated
            )
        return LinalgDecomposition(evd.values, phase_aligned_funcs)
